/***************************************************************************/
/*                                                                         */
/*                         Types for FKISS Extensions                      */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

#ifndef __FKISSTYPES_H
#define __FKISSTYPES_H

#include "KISSTypes.h"

#define FKISS_ENABLE "EventHandler"

#define MAX_FKISS_TIMER     512
#define MAX_FKISS_LABEL     512

/*
 * Forward declaration of event list type - fully declared later.
 */

typedef  struct __fks_event *FKISS_EVENT_LIST;

#define MaxIdentifierLen 32

/***************************************************************************/
/*
 * Types required for definition of cel groups.
 */

typedef struct __celarr {
                          KISS_CEL_LIST *cels;
                          int           numCel;
                          int           capacity;
                        } CEL_ARRAY_STR;

typedef CEL_ARRAY_STR *CEL_ARRAY_PTR;

typedef struct __frame {
                         int             frameNo;
                         CEL_ARRAY_STR   cels;
                         struct __frame  *nextFrame;
                       } FKISS_FRAME_TYPE;

typedef FKISS_FRAME_TYPE *FKISS_FRAME_LIST;

typedef struct __celgrp {
                          char             name[MaxIdentifierLen];
                          CEL_ARRAY_STR    cels;
                          FKISS_FRAME_LIST frames;
                          int              maxFrame;
                          int              currentFrame;
                          struct __celgrp  *nextGroup;
                        } FKISS_GROUP_TYPE;

typedef FKISS_GROUP_TYPE *FKISS_GROUP_LIST;

/***************************************************************************/
/*
 * New variable type for FKISS3, allowing any number of variable, with
 * better symbolic names than the original spec allowed.
 */

typedef struct __vartyp {
                           char             name[MaxIdentifierLen];
                           int              flags;
                           int              value;
                           FKISS_EVENT_LIST label;
                           FKISS_EVENT_LIST alarm;
                           struct __vartyp  *next;
                        } FKISS_VARIABLE_TYPE;

typedef FKISS_VARIABLE_TYPE *FKISS_VARIABLE_LIST;

#define VARIABLE_K_VALUE   (1u<<0)
#define VARIABLE_K_LABEL   (1u<<1)
#define VARIABLE_K_ALARM   (1u<<2)

#define VARIABLE_K_TYPEMASK (VARIABLE_K_LABEL | VARIABLE_K_VALUE | VARIABLE_K_ALARM)

/***************************************************************************/
/* Basic integer type for FKISS3 - could be a constant or a variable
 */

typedef struct __intval { int     flags;
                          union {
                            int                 constant;
                            FKISS_VARIABLE_LIST variable; } val; } FKISS_INTEGERVALUE;

#define FKINT_K_VARIABLE (1u<<0)
#define FKINT_K_CONSTANT 0
#define FKINT_K_TYPE     0xffffffff

/***************************************************************************/

typedef enum { fic_unknown,
               fic_cel,
               fic_objptr,
               fic_objvar,
               fic_group } FKISS_INSTANCE_CLASS;

typedef struct {
                 FKISS_INSTANCE_CLASS type;
                 void                *addr;
               } FKISS_INSTANCE_TYPE;

/***************************************************************************/

typedef struct __fksmove { FKISS_INSTANCE_TYPE object;
                           FKISS_INTEGERVALUE  x;
                           FKISS_INTEGERVALUE  y; } FKISS_MOVE;

typedef struct __fksmvrnd { FKISS_INSTANCE_TYPE object;
                            FKISS_INTEGERVALUE  mindist;
                            FKISS_INTEGERVALUE  maxdist; } FKISS_MOVERAND;

typedef struct __fksmvby { FKISS_INSTANCE_TYPE object1;
                           FKISS_INSTANCE_TYPE object2;
                           FKISS_INTEGERVALUE  dist;  } FKISS_MOVEBY;

typedef struct __transp  { FKISS_INSTANCE_TYPE instance;
                           FKISS_INTEGERVALUE  dtransp; } FKISS_TRANSPARENT;

typedef struct __fkcoord { FKISS_INTEGERVALUE x;
                           FKISS_INTEGERVALUE y; } FKISS_COORDS;

typedef struct __fkstimer { FKISS_INTEGERVALUE alarm;
                            FKISS_INTEGERVALUE delay;
                            FKISS_INTEGERVALUE variance; } FKISS_TIMER;

typedef struct __fksif    { FKISS_INSTANCE_TYPE instance;
                            FKISS_INTEGERVALUE  alarm;
                            FKISS_INTEGERVALUE  delay; } FKISS_IFTIMER;

typedef struct __setfix { FKISS_INSTANCE_TYPE obj;
                          FKISS_INTEGERVALUE  fix; } FKISS_SETFIX;

/***************************************************************************/
/* FKISS3 arithmetic structures
 */

typedef struct { FKISS_VARIABLE_LIST var;  /* LHS must be a variable, not a constant */
                 FKISS_INTEGERVALUE  op1;
                 FKISS_INTEGERVALUE  op2; } FKISS_ARITHMETICOP;

/* Object-based arithmetics */

typedef struct { FKISS_VARIABLE_LIST var;
                 FKISS_INSTANCE_TYPE obj1;
                 FKISS_INSTANCE_TYPE obj2; } FKISS_LETOBJECT;

/* Cel-based arithmetics */

typedef struct { FKISS_VARIABLE_LIST var;
                 FKISS_INSTANCE_TYPE cel1;
                 FKISS_INSTANCE_TYPE cel2; } FKISS_LETCEL;

/* Assignments not requiring any parameters */

typedef struct { FKISS_VARIABLE_LIST var; } FKISS_LETSPECIAL;

/* Assignments based on other things! */

typedef struct {
                 FKISS_VARIABLE_LIST var;
                 FKISS_INTEGERVALUE  timer;
               } FKISS_LETTIMER;


typedef struct {
                 FKISS_VARIABLE_LIST var;
                 FKISS_INSTANCE_TYPE group;
               } FKISS_LETGROUP;


/***************************************************************************/
/* FKISS3 logical test structures */

typedef struct { FKISS_INTEGERVALUE val1;
                 FKISS_INTEGERVALUE val2; } FKISS_LOGICALIF;

typedef enum { __notif, __elsepending, __endifpending } FKISS_LOGICALSTATE;

/***************************************************************************/
/* FKISS3 goto/gosub structures */

typedef struct { FKISS_INTEGERVALUE prob;
                 FKISS_INTEGERVALUE label1;
                 FKISS_INTEGERVALUE label2; } FKISS_BRANCH;

/***************************************************************************/
/* FKISS4 structures; */


typedef struct { FKISS_VARIABLE_LIST var;
                 char                *keylist; } FKISS_LETKEY;

typedef struct { FKISS_INSTANCE_TYPE child;
                 FKISS_INSTANCE_TYPE parent; } FKISS_ATTACH;

typedef struct { FKISS_INSTANCE_TYPE object;
                 FKISS_INTEGERVALUE  min;
                 FKISS_INTEGERVALUE  max; } FKISS_RESTRICT;

typedef struct { FKISS_INSTANCE_TYPE group;
                 FKISS_INTEGERVALUE  frameno; } FKISS_SETFRAME;

typedef struct {
                 FKISS_INTEGERVALUE  label;
                 FKISS_INTEGERVALUE  loopcount;
                 FKISS_VARIABLE_LIST var;
                 int                 upper;  /* Actual upper limit */
                 int                 count;  /* Actual loop counter */
               } FKISS_REPEAT;

typedef struct { FKISS_VARIABLE_LIST var;
                 FKISS_INSTANCE_TYPE inst; } FKISS_LETINSTANCE;

typedef struct { FKISS_VARIABLE_LIST var;
                 FKISS_INSTANCE_TYPE inst1;
                 FKISS_INSTANCE_TYPE inst2; } FKISS_LETINSTANCE2;

typedef struct { FKISS_INSTANCE_TYPE inst;
                 FKISS_INTEGERVALUE  palno; } FKISS_SETKCF;

typedef struct { FKISS_VARIABLE_LIST var;
                 char                *key; } FKISS_LOADVALUE;

typedef struct { char                *key;
                 FKISS_INTEGERVALUE  val; } FKISS_SAVEVALUE;

/***************************************************************************/

typedef enum { fks_map,            fks_unmap,
               fks_altmap,         fks_sound,
               fks_move,           fks_changeset,
               fks_changecol,      fks_timer,
               fks_randomtimer,    fks_nop,
               fks_quit,           fks_music,
               fks_debug,          fks_iffixed,
               fks_ifnotfixed,     fks_ifmapped,
               fks_ifnotmapped,    fks_ifmoved,
               fks_ifnotmoved,     fks_setfix,
               fks_transparent,    fks_viewport,
               fks_window,         fks_moveto,
               fks_movebyx,        fks_movebyy,
               fks_moverandx,      fks_moverandy,
               fks_movetorand,     fks_notify,
               fks_shell,

              /* FKISS 3 & 4 extensions */

               fks_let,            fks_letadd,
               fks_letsub,         fks_letmul,
               fks_letdiv,         fks_letmod,
               fks_letrandom,      fks_letobjectx,
               fks_letobjecty,     fks_letfix,
               fks_letmapped,      fks_letset,
               fks_letpal,         fks_letmousex,
               fks_letmousey,      fks_letcatch,
               fks_letcollide,     fks_letinside,
               fks_lettransp,      fks_lifeq,
               fks_lifne,          fks_lifgt,
               fks_liflt,          fks_else,
               fks_elseifeq,       fks_elseifne,
               fks_elseifgt,       fks_elseiflt,
               fks_endif,          fks_break,
               fks_goto,           fks_gosub,
               fks_gotorandom,     fks_gosubrandom,
               fks_ghost,

               /* FKISS4 extensions */

               fks_attach,         fks_glue,
               fks_detach,         fks_letparent,
               fks_letchild,       fks_letsibling,
               fks_letkey,         fks_letkeymap,

               /* FKISS 4b extensions */

               fks_lettimer,       fks_letunghost,
               fks_letinitx,       fks_letinity,
               fks_letwidth,       fks_letheight,
               fks_letframe,       fks_setframe,
               fks_restrictx,      fks_restricty,
               fks_exitloop,       fks_repeat,
               fks_letkcf,         fks_setkcf,
               fks_letdepth,       fks_savevalue,
               fks_loadvalue,      fks_deletevalue,
               fks_valuepool

              } FKISS_ACTION_ID;

#define FIRST_LOGICAL_IF fks_lifeq
#define LAST_LOGICAL_IF  fks_endif

typedef union {
                FKISS_INSTANCE_TYPE map;        /* Plus altmap, unmap */
                FKISS_INTEGERVALUE  changecol;
                FKISS_INTEGERVALUE  changeset;
                char                *debugmsg;   /* Debug & notify */
                char                *shellcmnd;  /* shell */
                FKISS_IFTIMER       iftimer;     /* iffixed, ifmapped, ifmoved + not */
                FKISS_MOVE          move;        /* +moveto */
                FKISS_MOVERAND      moverand;    /* moverandx & moverandy */
                FKISS_MOVEBY        moveby;      /* movebyx & movebyy */
                FKISS_INSTANCE_TYPE movetorand;
                char                *sound;      /* +music */
                FKISS_TIMER         timer;       /* timer & randomtimer */
                FKISS_SETFIX        setfix;
                FKISS_TRANSPARENT   transparent;
                FKISS_TRANSPARENT   ghost;        /* Ghost function uses same structure as transparent */
                FKISS_COORDS        window;       /* windowsize/viewport */
                FKISS_ARITHMETICOP  arithmetic;
                FKISS_LETOBJECT     letobject;
                FKISS_LETCEL        letcel;
                FKISS_LETSPECIAL    letspecial;
                FKISS_BRANCH        branch;
                FKISS_LOGICALIF     logicaltest;
                FKISS_INTEGERVALUE  label;
                FKISS_ATTACH        attach;       /* Plus glue */
                FKISS_INSTANCE_TYPE detach;
                FKISS_LETKEY        key;          /* letkey & letkeymap */
                FKISS_LETTIMER      lettimer;
                FKISS_RESTRICT      restrictarea;
                FKISS_LETGROUP      letgroup;
                FKISS_SETFRAME      setframe;
                FKISS_REPEAT        repeat;
                FKISS_LETINSTANCE   letinst;
                FKISS_LETINSTANCE2  letinst2;
                FKISS_SETKCF        setkcf;
                FKISS_LOADVALUE     loadvalue;
                FKISS_SAVEVALUE     savevalue;
                char                *poolkey;   /* deletevalue & poolname */
              } FKISS_ACTION_DETAILS;

typedef struct __fks_action
               { FKISS_ACTION_ID      fid;
                 FKISS_ACTION_DETAILS data;
                 int                  lineno;
                 struct __fks_action *next_action;
               } FKISS_ACTION_STR;

typedef FKISS_ACTION_STR *FKISS_ACTION_LIST;

/***************************************************************************/

typedef enum {
               fks_initialize = 0x00010000,
               fks_begin,       fks_end,
               fks_alarm,       fks_set,
               fks_col,         fks_never,
               fks_in,          fks_out,
               fks_stillin,     fks_stillout,
               fks_version,     fks_collide,
               fks_apart,       fks_label,
               fks_overflow,    fks_detached,
               fks_key,         fks_keyrelease,
               fks_press,       fks_catch,
               fks_fixcatch,    fks_release,
               fks_drop,        fks_fixdrop,
               fks_unfix,

               /* FKISS4 */

               fks_mouseout,    fks_mousein   /* Order of mouse event types is trigger order */

                } FKISS_EVENT_ID;

typedef struct __fks_alarm { FKISS_INTEGERVALUE alarm;
                             int                state;
                           } FKISS_ALARM_EVENT;

typedef struct __fks_touch { FKISS_INSTANCE_TYPE object1;
                             FKISS_INSTANCE_TYPE object2; } FKISS_TOUCH_EVENT;

typedef struct __fks_collide { FKISS_INSTANCE_TYPE cel1;
                               FKISS_INSTANCE_TYPE cel2; } FKISS_COLLIDE_EVENT;

typedef union {
                FKISS_ALARM_EVENT   alarm;
                FKISS_TOUCH_EVENT   touch;
                FKISS_COLLIDE_EVENT collide;
                FKISS_INTEGERVALUE  palette;
                FKISS_INTEGERVALUE  set;
                FKISS_INSTANCE_TYPE mouse;
                FKISS_INTEGERVALUE  version;
                FKISS_INTEGERVALUE  label;
                char                *keyname;   /* Used only during parsing */
                int                 key;
                FKISS_INSTANCE_TYPE detached;
                FKISS_INSTANCE_TYPE pointer; } FKISS_EVENT_DETAILS;

typedef struct __fks_event
               { FKISS_EVENT_ID      fid;
                 FKISS_EVENT_DETAILS data;
                 FKISS_ACTION_LIST   actions;
                 int                 lineno;
                 struct __fks_event  *next_event;
                } FKISS_EVENT_STR;

#define FKISS_K_Enabled    (1u<<0)
#define FKISS_K_Active     (1u<<1)
#define FKISS_K_Handler    (1u<<2)

#define FKISS_K_Initial  FKISS_K_Enabled

#endif
