/***************************************************************************/
/*                                                                         */
/*                              FKISS function parser                      */
/*                                                                         */
/***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include "assert.h"

#include "BOOL.h"

#include "KISSTypes.h"
#include "FKISSTypes.h"
#include "KISS_data.h"
#include "readkiss.h"
#include "utilities.h"
#include "message.h"
#include "keyhandler.h"
#include "celgroup.h"
#include "poolio.h"
#include "FKParamTypes.h"

static FKISS_ACTION_LIST *currentevent=NULL, lastaction=NULL;

/***************************************************************************/

#define FKS_P_Success      0

#define FKS_P_BadObj       1
#define FKS_P_ObjExp       2
#define FKS_P_CelExp       3
#define FKS_P_BadCel       4
#define FKS_P_ItemOrCel    5
#define FKS_P_BadSet       6
#define FKS_P_BadAlarm     7
#define FKS_P_BadPalette   8
#define FKS_P_BadSyntax    9
#define FKS_P_BadFunc     10
#define FKS_P_BadFile     11
#define FKS_P_BadInt      12
#define FKS_P_BadParams   13
#define FKS_P_BadString   14
#define FKS_P_BadVar      15
#define FKS_P_BadLabel    16
#define FKS_P_Ignored    -17
#define FKS_P_MissFunc   -18
#define FKS_P_UnclosedIf -19
#define FKS_P_OrphanIf   -20
#define FKS_P_NestedIf   -21
#define FKS_P_BadKeyName  22
#define FKS_P_DupHandler  23
#define FKS_P_FileTooLong 24
#define FKS_P_AmbigCel    25
#define FKS_P_UnexpGrp    33
#define FKS_P_BadGrp      34
#define FKS_P_BadEvent    35
#define FKS_P_BadTag      36

/* Variables to record current state of logical if blocks */

#define IfState_None       0
#define IfState_IfPart     1
#define IfState_ElsePart   2

static int LogicalIfState = IfState_None;

static int begin_logical_if_line = -1;

static BOOL skip_to_next_event = FALSE;

extern int line_count;  /* From KISS_IO.h */

static char **fkiss_text_ptr, *fkiss_start_ptr;
static char *saved_fkiss_text=NULL;
static int savedTextLen = 0;

/***************************************************************************/

static void save_fkiss_text(char **ptr)
{
  int len;
  fkiss_text_ptr  = ptr;
  fkiss_start_ptr = *ptr;
  len = strlen(fkiss_start_ptr) + 1;
  if ( len > savedTextLen )
  {
    saved_fkiss_text = realloc(saved_fkiss_text,len);
    savedTextLen = len;
  }
  strcpy(saved_fkiss_text,fkiss_start_ptr);
}

/***************************************************************************/

static void fkiss_report(int errcode)
{
  char msgid[10];
  int srclen = (int)*fkiss_text_ptr - (int)fkiss_start_ptr;

  saved_fkiss_text[srclen] = 0;

  sprintf(msgid,"FKS%1d",errcode>0 ? errcode : -errcode);
  message_error(msgid,saved_fkiss_text,begin_logical_if_line);
}

/***************************************************************************/

static FKISS_VARIABLE_LIST read_variable(char *param,int flags)
{
  if ( isalpha(*param) )
  {
    FKISS_VARIABLE_LIST var=fkiss_variables, prev = NULL;

    while ( var )
    {
      if ( struncmp(var->name,param,MaxIdentifierLen) == 0 )
        break;
      prev = var;
      var = var->next;
    }

    if ( !var )
    {
      var = safe_alloc(sizeof(FKISS_VARIABLE_TYPE));
      memset(var,0,sizeof(FKISS_VARIABLE_TYPE));
      var->next  = NULL;
      var->value = 0;
      var->label = NULL;
      var->flags = 0;
      strncpy(var->name,param,MaxIdentifierLen);
      if ( prev )
        prev->next = var;
      else fkiss_variables = var;
    }
    var->flags |= flags;
    return var;
  }
  return NULL;
}

/***************************************************************************/

static char *read_string(char *param)
{ char buff[MAX_INPUT_BUFFER_LEN],*p=buff;
  char *s;

  if ( *param != '\"' ) return NULL;
  param++;
  while ( *param && (*param != '\"') )
  {
    if ( *param == '\\' )
    {
      param++;
      switch (*param)
      {
        case 't' : *p = '\t'; break;
        case 'n' : *p = '\n'; break;
        case 'r' : *p = '\r'; break;
        default  : *p = *param; break;
      }
    }
    else *p = *param;
    p++;
    param++;
  }
  if ( !*param ) return FALSE;

  *p = 0;
  s = safe_alloc(strlen(buff)+1);
  strcpy(s,buff);
  return s;
}

/***************************************************************************/

static int read_keyid(char *param)
{
  char *s;
  int keyid;

  s = read_string(param);
  if ( !s )
    return -1;

  keyid = keyhandler_keynum(s);
  free(s);

  return keyid;
}

/***************************************************************************/

#define COG_OBJ  (1u<<0)
#define COG_CEL  (1u<<1)
#define COG_VAR  (1u<<2)
#define COG_GRP  (1u<<3)
#define COG_ANY  (COG_OBJ | COG_CEL | COG_VAR | COG_GRP)

static int read_instance(char *param,int tmask,FKISS_INSTANCE_TYPE *instance)
{ int temp;
  char celname[MAXLEAFNAMELEN+1];
  KISS_OBJECT_LIST obj;
  KISS_CEL_LIST cel;
  FKISS_VARIABLE_LIST variable;
  FKISS_GROUP_LIST group;

  instance->type = fic_unknown;

  switch ( *param )
  {
    case '\"' :
       if ( !(tmask & COG_CEL) )
         return FKS_P_ObjExp;

       temp = strlen(param)-2;

       if ( temp > MAXLEAFNAMELEN )
         temp = MAXLEAFNAMELEN;

       strncpy(celname,&(param[1]),temp);
       celname[temp] = 0;
       cel = get_cel_descriptor(celname);
       if ( !cel )
         return FKS_P_BadCel;
       if ( cel->duplicate )
         return FKS_P_AmbigCel;

       instance->type = fic_cel;
       instance->addr = cel;
       break;

    case '#' :
       if ( !(tmask & COG_OBJ) )
         return FKS_P_CelExp;
       if ( !convert_to_int(&(param[1]),&temp) )
         return FKS_P_BadObj;
       obj = get_object_mark(temp);
       if ( !obj )
         return FKS_P_BadObj;
       instance->type = fic_objptr;
       instance->addr = obj;
       break;

    case '!' :
       if ( !(tmask & COG_GRP) )
         return FKS_P_UnexpGrp;

       /*
        * Architecturally, this shouldn't really be here, but it
        * simplifies the fkiss function array considerably.
        */

       KISS_Version = MAX(KISS_Version,KISS_V_FKISS4);

       group = celgroup_findGroup(&(param[1]));
       if ( !group )
         return FKS_P_BadGrp;
       instance->type = fic_group;
       instance->addr = group;
       break;

    default :
       if ( !(tmask & COG_VAR) )
         return FKS_P_ItemOrCel;

       variable = read_variable(param,VARIABLE_K_VALUE);
       if ( variable )
       {
         /*
          * See architectural note for cel groups above.
          */

         KISS_Version = MAX(KISS_Version,KISS_V_FKISS3);

         instance->type = fic_objvar;
         instance->addr = variable;
         break;
       }
       return FKS_P_ItemOrCel;
  }
  return FKS_P_Success;
}

#define read_cel(p,v)          read_instance(p,COG_CEL,v)
#define read_celgrp(p,v)       read_instance(p,COG_CEL | COG_GRP,v)
#define read_object(p,v)       read_instance(p,COG_OBJ,v)
#define read_objorvar(p,v)     read_instance(p,COG_OBJ | COG_VAR,v)
#define read_cog(p,v)          read_instance(p,COG_OBJ | COG_CEL | COG_GRP,v)
#define read_group(p,v)        read_instance(p,COG_CEL | COG_GRP,v)
#define read_cogv(p,v)         read_instance(p,COG_ANY,v)

/***************************************************************************/

void fkiss_error(char *msgid,char *rtn)
{
  message_report("Internal error : %s in routine %s\n",
             message_msgtext(msgid),rtn);
  exit(1);
}

/***************************************************************************/

static BOOL sameInstance(FKISS_INSTANCE_TYPE i1,
                         FKISS_INSTANCE_TYPE i2)
{
  return (i1.type == i2.type) && (i1.addr == i2.addr);
}

/***************************************************************************/

static int add_event(int fid,
                      FKISS_EVENT_DETAILS *event,
                      FKISS_EVENT_LIST *head)
{
  FKISS_EVENT_LIST newevent=NULL,e;
  BOOL duplicate = FALSE;
  int status = FKS_P_Success;

  /*
   * Check for any unclosed IF from the previous event
   */

  if ( LogicalIfState != IfState_None )
    fkiss_report(FKS_P_UnclosedIf);

  LogicalIfState = IfState_None;

  /* Ensure that if we have two identical events, the action lists are
     concatenated
   */

  if ( head )
  {
    e = *head;
    while ( e && !duplicate )
    {
      if ( e->fid == fid )
      {
        switch ( fid )
        {
          case fks_initialize :
          case fks_begin :
          case fks_end :
          case fks_never :
          case fks_overflow :
                 duplicate = TRUE; break;

          case fks_set :
          case fks_col :
          case fks_version :
                 duplicate = (e->data.version.val.constant == event->version.val.constant);
                 break;

          case fks_alarm :
                 duplicate = ( ((e->data.alarm.alarm.flags & FKINT_K_TYPE) ==
                                (event->alarm.alarm.flags & FKINT_K_TYPE) ) &&
                               (e->data.alarm.alarm.val.constant ==
                                   event->alarm.alarm.val.constant) );
                 break;

          case fks_in :
          case fks_out :
          case fks_stillin :
          case fks_stillout :
                 duplicate = (sameInstance(e->data.touch.object1,event->touch.object1) &&
                               sameInstance(e->data.touch.object2,event->touch.object2)) ||
                             (sameInstance(e->data.touch.object1,event->touch.object2) &&
                               sameInstance(e->data.touch.object2,event->touch.object1));
                 break;

          case fks_apart :
          case fks_collide :
                 duplicate = (sameInstance(e->data.collide.cel1,event->collide.cel1) &&
                               sameInstance(e->data.collide.cel2,event->collide.cel2)) ||
                             (sameInstance(e->data.collide.cel1,event->collide.cel2) &&
                               sameInstance(e->data.collide.cel2,event->collide.cel1));
                 break;

          case fks_press :
          case fks_catch :
          case fks_fixcatch :
          case fks_release :
          case fks_drop :
          case fks_fixdrop :
          case fks_unfix :
                 duplicate = sameInstance(e->data.mouse,event->mouse);
                 break;

          case fks_label :
                 duplicate = ( ((e->data.label.flags & FKINT_K_TYPE) ==
                                  (event->label.flags & FKINT_K_TYPE) ) &&
                                (e->data.label.val.constant ==
                                   event->label.val.constant) );
                 break;

          case fks_key :
          case fks_keyrelease :
                 duplicate = (e->data.key == event->key);
                 break;

          case fks_detached :
                 duplicate = sameInstance(e->data.detached,event->detached);
                 break;

          case fks_mousein :
          case fks_mouseout :
                 duplicate = sameInstance(e->data.pointer,event->pointer);
                 break;

          default : fkiss_error("FKP4","add_event");

        }  /* Switch */
      }    /* event ids match */

      if ( !duplicate )
        e = e->next_event;
    }
  }

  if ( duplicate )
  {
    currentevent = &(e->actions);
    lastaction = *currentevent;
    while ( lastaction && lastaction->next_action )
      lastaction = lastaction->next_action;
    status = FKS_P_DupHandler;
  }
  else
  {
    FKISS_EVENT_LIST last = *head;

    newevent = safe_alloc(sizeof(FKISS_EVENT_STR));
    memcpy(&(newevent->data),event,sizeof(FKISS_EVENT_DETAILS));
    newevent->fid = (FKISS_EVENT_ID)fid;
    newevent->actions = NULL;
    newevent->next_event = NULL;

    if ( !last )
      *head = newevent;
    else
    {
      while ( last && last->next_event ) last = last->next_event;
      last->next_event = newevent;
    }

    currentevent = &(newevent->actions);
    newevent->lineno = line_count;

  }

  return status;
}

/***************************************************************************/

static int fkiss_createkeyevent(int funcno,
                                FKISS_EVENT_DETAILS *event)
{
  char *key = event->keyname;
  int keyid = keyhandler_keynum(key);
  if ( keyid < 0 )
    return FKS_P_BadKeyName;
  free(key);
  event->key = keyid;
  return add_event(funcno,event,&fkiss_key_list);
}

/***************************************************************************/

static int fkiss_createmiscevent(int funcno,
                                 FKISS_EVENT_DETAILS *event)
{
  return add_event(funcno,event,&fkiss_misc_list);
}

/***************************************************************************/

static void checkrange(int val,int min,int max,int errno)
{
  if ( (val < min) || (val>max) )
    fkiss_report(errno);
}

/***************************************************************************/

static int fkiss_createalarm(int funcno,
                             FKISS_EVENT_DETAILS *event)
{
  FKISS_INTEGERVALUE val = event->alarm.alarm;

  if ( (val.flags & FKINT_K_TYPE) == FKINT_K_CONSTANT )
  {
    checkrange(val.val.constant,0,MAX_FKISS_TIMER-1,FKS_P_BadAlarm);
  } else event->alarm.alarm.val.variable->flags |= VARIABLE_K_ALARM;

  return add_event(funcno,event,&fkiss_alarm_list);
}

/***************************************************************************/

static int fkiss_createlabel(int funcno,
                             FKISS_EVENT_DETAILS *event)
{
  FKISS_INTEGERVALUE val = event->label;

  if ( (val.flags & FKINT_K_TYPE) == FKINT_K_CONSTANT )
  {
    checkrange(val.val.constant,0,MAX_FKISS_LABEL-1,FKS_P_BadLabel);
  }
  else val.val.variable->flags |= VARIABLE_K_LABEL;

  return add_event(funcno,event,&fkiss_label_list);
}

/***************************************************************************/

static int fkiss_createsetevent(int funcno,
                                FKISS_EVENT_DETAILS *event)
{
  checkrange(event->set.val.constant,0,MAX_SETS-1,FKS_P_BadSet);

  return fkiss_createmiscevent(funcno,event);
}

/***************************************************************************/

static int fkiss_createcolevent(int funcno,
                                FKISS_EVENT_DETAILS *event)
{
  checkrange(event->palette.val.constant,0,PAL_MAX_GRP-1,FKS_P_BadPalette);
  return fkiss_createmiscevent(funcno,event);
}

/***************************************************************************/

static int fkiss_createmouseevent(int funcno,
                                  FKISS_EVENT_DETAILS *event)
{
  return add_event(funcno,event,&fkiss_mouse_list);
}

/***************************************************************************/

static int fkiss_createpointerevent(int funcno,
                                  FKISS_EVENT_DETAILS *event)
{
  return add_event(funcno,event,&fkiss_pointer_list);
}

/***************************************************************************/

static int fkiss_createcollideevent(int funcno,
                                    FKISS_EVENT_DETAILS *event)
{
  return add_event(funcno,event,&fkiss_collision_list);
}

/***************************************************************************/

static int fkiss_add_action(int funcno,FKISS_ACTION_DETAILS *action)
{
  FKISS_ACTION_LIST newact;
  int status = FKS_P_Success;

  if ( currentevent == NULL )
    return FKS_P_Ignored;

  if ( *currentevent == NULL )
    lastaction = NULL;

  /* Check for legality or otherwise of logical if statements */

  switch ( funcno )
  {
    case fks_lifne :
    case fks_lifeq :
    case fks_lifgt :
    case fks_liflt :
           if ( LogicalIfState != IfState_None )
             fkiss_report(FKS_P_NestedIf);
           LogicalIfState = IfState_IfPart;
           begin_logical_if_line = line_count;
           break;

    case fks_elseifne :
    case fks_elseifeq :
    case fks_elseifgt :
    case fks_elseiflt :
    case fks_else :
           if ( LogicalIfState != IfState_IfPart )
             fkiss_report(FKS_P_OrphanIf);

           LogicalIfState = (funcno == fks_else) ? IfState_ElsePart
                                                 : IfState_IfPart;
           break;

    case fks_endif :
           if ( LogicalIfState == IfState_None )
             fkiss_report(FKS_P_OrphanIf);

           LogicalIfState = IfState_None;
           break;
  }

  newact = safe_alloc(sizeof(FKISS_ACTION_STR));

  memcpy(&(newact->data),action,sizeof(FKISS_ACTION_DETAILS));
  newact->fid = (FKISS_ACTION_ID)funcno;
  newact->next_action = NULL;

  newact->lineno = line_count;

  if ( lastaction ) lastaction->next_action = newact;
  else *currentevent = newact;
  lastaction = newact;

  return status;
}

/***************************************************************************/
/*
 * Definition of all FKISS functions.
 * It is legal to have multiple entries for the same function (to distinguish
 * between different parameter sets) - in this case, the most restrictive
 * parameter set should go first.
 *
 * e.g. alarm appears twice, once with a constant parameter (basic FKISS) and
 * once with a "variable" (actually a symbol name) (FKISS 4).
 *
 * Note: Cel Groups are not distinguished in this way. Instead, the use of
 * a group automatically promotes the set to at least FKISS4 (see read_instance).
 * Similarly, /any/ use of a variable promotoes the set to at least FKISS3.
 */

FKISS_LOAD_STR fkiss_functions[] =
    {
      /* Events */

      { "alarm",      fks_alarm,       FVERS(0) | P_CONSTANT(0),                       fkiss_createalarm },
      { "alarm",      fks_alarm,       FVERS(3) | P_INTVAL(0),                         fkiss_createalarm },
      { "apart",      fks_apart,       FVERS(1) | P_CELGRP(0)|P_CELGRP(1),             fkiss_createcollideevent },
      { "begin",      fks_begin,       FVERS(0),                                       fkiss_createmiscevent },
      { "bigin",      fks_begin,       FVERS(0) | FERROR(FKISS_MISSPELLING),           fkiss_createmiscevent }, /* Common misspelling */
      { "catch",      fks_catch,       FVERS(0) | P_COG(0) | FERROR(FKISS_NOTPORTABLE),fkiss_createmouseevent },
      { "col",        fks_col,         FVERS(0) | P_CONSTANT(0),                       fkiss_createcolevent },
      { "collide",    fks_collide,     FVERS(1) | P_CELGRP(0)|P_CELGRP(1),             fkiss_createcollideevent },
      { "collide",    fks_collide,     FVERS(3) | P_COG(0)|P_COG(1),                   fkiss_createcollideevent },
      { "detached",   fks_detached,    FVERS(3) | P_OBJECT(0),                         fkiss_createmiscevent },
      { "drop",       fks_drop,        FVERS(0) | P_COG(0) | FERROR(FKISS_NOTPORTABLE),fkiss_createmouseevent},
      { "end",        fks_end,         FVERS(0),                                       fkiss_createmiscevent },
      { "fixcatch",   fks_fixcatch,    FVERS(0) | P_COG(0) | FERROR(FKISS_NOTPORTABLE),fkiss_createmouseevent },
      { "fixdrop",    fks_fixdrop,     FVERS(0) | P_COG(0) | FERROR(FKISS_NOTPORTABLE),fkiss_createmouseevent },
      { "in",         fks_in,          FVERS(1) | P_OBJECT(0)|P_OBJECT(1),             fkiss_createcollideevent },
      { "in",         fks_in,          FVERS(3) | P_COG(0)|P_COG(1),                   fkiss_createcollideevent },
      { "initialize", fks_initialize,  FVERS(0),                                       fkiss_createmiscevent },
      { "initialise", fks_initialize,  FVERS(0) | FERROR(FKISS_MISSPELLING),           fkiss_createmiscevent },
      { "keypress",   fks_key,         FVERS(3) | P_STRING(0),                         fkiss_createkeyevent },
      { "keyrelease", fks_keyrelease,  FVERS(3) | P_STRING(0),                         fkiss_createkeyevent },
      { "label",      fks_label,       FVERS(2) | P_CONSTANT(0),                       fkiss_createlabel },
      { "label",      fks_label,       FVERS(3) | P_INTVAL(0),                         fkiss_createlabel },
      { "mousein",    fks_mousein,     FVERS(3) | P_COG(0),                            fkiss_createpointerevent },
      { "mouseout",   fks_mouseout,    FVERS(3) | P_COG(0),                            fkiss_createpointerevent },
      { "never",      fks_never,       FVERS(0),                                       fkiss_createmiscevent },
      { "nothing",    fks_never,       FVERS(0) | FERROR(FKISS_SYNONYM),               fkiss_createmiscevent },  /* Synonym for never */
      { "overflow",   fks_overflow,    FVERS(2),                                       fkiss_createmiscevent },
      { "out",        fks_out,         FVERS(1) | P_OBJECT(0)|P_OBJECT(1),             fkiss_createcollideevent },
      { "out",        fks_out,         FVERS(3) | P_COG(0)|P_COG(1),                   fkiss_createcollideevent },
      { "press",      fks_press,       FVERS(0) | P_COG(0),                            fkiss_createmouseevent },
      { "release",    fks_release,     FVERS(0) | P_COG(0),                            fkiss_createmouseevent },
      { "set",        fks_set,         FVERS(0) | P_CONSTANT(0),                       fkiss_createsetevent },
      { "stillin",    fks_stillin,     FVERS(1) | P_OBJECT(0)|P_OBJECT(1),             fkiss_createcollideevent },
      { "stillin",    fks_stillin,     FVERS(3) | P_COG(0)|P_COG(1),                   fkiss_createcollideevent },
      { "stillout",   fks_stillout,    FVERS(1) | P_OBJECT(0)|P_OBJECT(1),             fkiss_createcollideevent },
      { "stillout",   fks_stillout,    FVERS(3) | P_COG(0)|P_COG(1),                   fkiss_createcollideevent },
      { "unfix",      fks_unfix,       FVERS(0) | P_COG(0),                            fkiss_createmouseevent },
      { "version",    fks_version,     FVERS(1) | P_CONSTANT(0),                       fkiss_createmiscevent },

     /* Actions */

      { "add",        fks_letadd,       FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },
      { "altmap",     fks_altmap,       FVERS(0) | P_COG(0),                              NULL },
      { "altmap",     fks_altmap,       FVERS(3) | P_COGV(0),                             NULL },
      { "attach",     fks_attach,       FVERS(3) | P_OBJVAR(0)|P_OBJVAR(1),               NULL },

      { "changecol",  fks_changecol,    FVERS(0) | P_INTVAL(0),                           NULL },
      { "changeset",  fks_changeset,    FVERS(0) | P_INTVAL(0),                           NULL },

      { "debug",      fks_debug,        FVERS(0) | P_STRING(0),                           NULL },
      { "deletevalue",fks_deletevalue,  FVERS(3) | P_IDTAG(0),                            NULL },
      { "detach",     fks_detach,       FVERS(3) | P_OBJVAR(0),                           NULL },
      { "div",        fks_letdiv,       FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },

      { "else",       fks_else,         FVERS(2),                                         NULL },
      { "elseifequal",fks_elseifeq,     FVERS(3) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "elseifgreaterthan", fks_elseifgt,
                                        FVERS(3) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "elseiflessthan", fks_elseiflt, FVERS(3) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "elseifnotequal", fks_elseifne, FVERS(3) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "endif",      fks_endif,        FVERS(2),                                         NULL },
      { "exitevent",  fks_break,        FVERS(2),                                         NULL },
      { "exitrepeat", fks_exitloop,     FVERS(3),                                         NULL },

      { "ghost",      fks_ghost,        FVERS(2) | P_COG(0)|P_INTVAL(1),                  NULL },
      { "ghost",      fks_ghost,        FVERS(3) | P_COGV(0)|P_INTVAL(1),                 NULL },
      { "glue",       fks_glue,         FVERS(3) | P_OBJVAR(0)|P_OBJVAR(1),               NULL },
      { "goto",       fks_goto,         FVERS(2) | P_INTVAL(0),                           NULL },
      { "gotorandom", fks_gotorandom,   FVERS(2) | P_INTVAL(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "gosub",      fks_gosub,        FVERS(2) | P_INTVAL(0),                           NULL },
      { "gosubrandom",fks_gosubrandom,  FVERS(2) | P_INTVAL(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },

      { "ifequal",    fks_lifeq,        FVERS(2) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "iffixed",    fks_iffixed,      FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "iffixed",    fks_iffixed,      FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifgreaterthan", fks_lifgt,     FVERS(2) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "iflessthan", fks_liflt,        FVERS(2) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "ifmapped",   fks_ifmapped,     FVERS(1) | P_CELGRP(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifmoved",    fks_ifmoved,      FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifmoved",    fks_ifmoved,      FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifnotequal", fks_lifne,        FVERS(2) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "ifnotfixed", fks_ifnotfixed,   FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifnotfixed", fks_ifnotfixed,   FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifnotmapped",fks_ifnotmapped,  FVERS(1) | P_CELGRP(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifnotmapped", fks_ifnotmapped, FVERS(3) | P_COGV(0)|P_INTVAL(1)|P_INTVAL(2),     NULL },
      { "ifnotmoved", fks_ifnotmoved,   FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "ifnotmoved", fks_ifnotmoved,   FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },

      { "let",        fks_let,          FVERS(2) | P_VARIABLE(0)|P_INTVAL(1),             NULL },
      { "letcatch",   fks_letcatch,     FVERS(2) | P_VARIABLE(0),                         NULL },
      { "letchild",   fks_letchild,     FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letcollide", fks_letcollide,   FVERS(2) | P_VARIABLE(0)|P_CELGRP(1)|P_CELGRP(2), NULL },
      { "letcollide", fks_letcollide,   FVERS(3) | P_VARIABLE(0)|P_COGV(1)|P_COGV(2),     NULL },
      { "letdepth",   fks_letdepth,     FVERS(3) | P_VARIABLE(0)|P_COGV(1),               NULL },
      { "letfix",     fks_letfix,       FVERS(2) | P_VARIABLE(0)|P_OBJECT(1),             NULL },
      { "letfix",     fks_letfix,       FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letframe",   fks_letframe,     FVERS(3) | P_VARIABLE(0)|P_GROUP(1),              NULL },
      { "letheight",  fks_letheight,    FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letinitx",   fks_letinitx,     FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letinity",   fks_letinity,     FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letinside",  fks_letinside,    FVERS(2) | P_VARIABLE(0)|P_OBJECT(1)|P_OBJECT(2), NULL },
      { "letinside",  fks_letinside,    FVERS(3) | P_VARIABLE(0)|P_COGV(1)|P_COGV(2),     NULL },
      { "letkcf",     fks_letkcf,       FVERS(3) | P_VARIABLE(0)|P_COGV(1),               NULL },
      { "letkey",     fks_letkey,       FVERS(3) | P_VARIABLE(0)|P_STRING(1),             NULL },
      { "letkeymap",  fks_letkeymap,    FVERS(3) | P_VARIABLE(0)|P_STRING(1),             NULL },
      { "letmapped",  fks_letmapped,    FVERS(2) | P_VARIABLE(0)|P_CELGRP(1),             NULL },
      { "letmapped",  fks_letmapped,    FVERS(3) | P_VARIABLE(0)|P_COGV(1),               NULL },
      { "letmousex",  fks_letmousex,    FVERS(2) | P_VARIABLE(0),                         NULL },
      { "letmousey",  fks_letmousey,    FVERS(2) | P_VARIABLE(0),                         NULL },
      { "letobjectx", fks_letobjectx,   FVERS(2) | P_VARIABLE(0)|P_OBJECT(1),             NULL },
      { "letobjectx", fks_letobjectx,   FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letobjecty", fks_letobjecty,   FVERS(2) | P_VARIABLE(0)|P_OBJECT(1),             NULL },
      { "letobjecty", fks_letobjecty,   FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letpal",     fks_letpal,       FVERS(2) | P_VARIABLE(0),                         NULL },
      { "letparent",  fks_letparent,    FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "letset",     fks_letset,       FVERS(2) | P_VARIABLE(0),                         NULL },
      { "letsibling", fks_letsibling,   FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "lettimer",   fks_lettimer,     FVERS(3) | P_VARIABLE(0)|P_INTVAL(1),             NULL },
      { "lettransparent",  fks_lettransp, FVERS(2) | P_VARIABLE(0)|P_CELGRP(1),           NULL },
      { "lettransparent", fks_lettransp, FVERS(3) | P_VARIABLE(0)|P_COGV(1),                  NULL },
      { "letwidth",   fks_letwidth,     FVERS(3) | P_VARIABLE(0)|P_OBJVAR(1),             NULL },
      { "loadvalue",  fks_loadvalue,    FVERS(3) | P_VARIABLE(0)|P_IDTAG(1),                 NULL },

      { "map",        fks_map,          FVERS(0) | P_COG(0),                              NULL },
      { "map",        fks_map,          FVERS(3) | P_COGV(0),                             NULL },
      { "mod",        fks_letmod,       FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },
      { "move",       fks_move,         FVERS(0) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "move",       fks_move,         FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moveto",     fks_moveto,       FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moveto",     fks_moveto,       FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moverandx",  fks_moverandx,    FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moverandx",  fks_moverandx,    FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moverandy",  fks_moverandy,    FVERS(1) | P_OBJECT(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "moverandy",  fks_moverandy,    FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "movebyx",    fks_movebyx,      FVERS(1) | P_OBJECT(0)|P_OBJVAR(1)|P_INTVAL(2),   NULL },
      { "movebyx",    fks_movebyx,      FVERS(3) | P_OBJVAR(0)|P_OBJVAR(1)|P_INTVAL(2),   NULL },
      { "movebyy",    fks_movebyy,      FVERS(1) | P_OBJECT(0)|P_OBJVAR(1)|P_INTVAL(2),   NULL },
      { "movebyy",    fks_movebyy,      FVERS(3) | P_OBJVAR(0)|P_OBJVAR(1)|P_INTVAL(2),   NULL },
      { "movetorand", fks_movetorand,   FVERS(1) | P_OBJECT(0),                           NULL },
      { "movetorand", fks_movetorand,   FVERS(3) | P_OBJVAR(0),                           NULL },
      { "mul",        fks_letmul,       FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },
      { "music",      fks_music,        FVERS(1) | P_FILE(0),                             NULL },

      { "nop",        fks_nop,          FVERS(0),                                         NULL },
      { "notify",     fks_notify,       FVERS(1) | P_STRING(0),                           NULL },

      { "quit",       fks_quit,         FVERS(0),                                         NULL },

      { "random",     fks_letrandom,    FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },
      { "randomtimer",fks_randomtimer,  FVERS(0) | P_CONSTANT(0)|P_INTVAL(1)|P_INTVAL(2), NULL },
      { "randomtimer",fks_randomtimer,  FVERS(3) | P_INTVAL(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "repeat",     fks_repeat,       FVERS(3) | P_INTVAL(0)|P_INTVAL(1)|P_VARIABLE(2), NULL },
      { "restrictx",  fks_restrictx,    FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },
      { "restricty",  fks_restricty,    FVERS(3) | P_OBJVAR(0)|P_INTVAL(1)|P_INTVAL(2),   NULL },

      { "savevalue",  fks_savevalue,    FVERS(3) | P_IDTAG(0)|P_INTVAL(1),                NULL },
      { "setfix",     fks_setfix,       FVERS(1) | P_OBJECT(0)|P_INTVAL(1),               NULL },
      { "setfix",     fks_setfix,       FVERS(3) | P_OBJVAR(0)|P_INTVAL(1),               NULL },
      { "setframe",   fks_setframe,     FVERS(3) | P_GROUP(0)|P_INTVAL(1),                NULL },
      { "setkcf",     fks_setkcf,       FVERS(3) | P_COGV(0)|P_INTVAL(1),                 NULL },
      { "shell",      fks_shell,        FVERS(1) | P_STRING(0)| FERROR(FKISS_NOTPORTABLE),NULL },
      { "sound",      fks_sound,        FVERS(0) | P_FILE(0),                             NULL },
      { "sub",        fks_letsub,       FVERS(2) | P_VARIABLE(0)|P_INTVAL(1)|P_INTVAL(2), NULL },

      { "timer",      fks_timer,        FVERS(0) | P_CONSTANT(0)|P_INTVAL(1),             NULL },
      { "timer",      fks_timer,        FVERS(3) | P_INTVAL(0)|P_INTVAL(1),               NULL },
      { "transparent",fks_transparent,  FVERS(1) | P_COG(0)|P_INTVAL(1),                  NULL },
      { "transparent",fks_transparent,  FVERS(3) | P_COGV(0)|P_INTVAL(1),                 NULL },

      { "unmap",      fks_unmap,        FVERS(0) | P_COG(0),                              NULL },
      { "unmap",      fks_unmap,        FVERS(3) | P_COGV(0),                             NULL },

      { "valuepool",  fks_valuepool,    FVERS(3) | P_IDTAG(0),                               NULL },
      { "viewport",   fks_viewport,     FVERS(1) | P_INTVAL(0)|P_INTVAL(1),               NULL },

      { "windowsize", fks_window,       FVERS(1) | P_INTVAL(0)|P_INTVAL(1),               NULL },

      { NULL,         0,                0,                                                NULL }
    };

/***************************************************************************/

#define MAX_FUNCNAME_LEN 63

static int fkiss_function_no(int fid)
{ int fno=0;

  while ( fkiss_functions[fno].funcname )
    if ( fkiss_functions[fno].fid == fid )
      return fno;
    else fno++;

  fprintf(stderr,"Bad fkiss function id %1d",fid);
  exit(1);
  return -1;
}

/***************************************************************************/

char *fkiss_function_name(int fid)
{ int fno = fkiss_function_no(fid);

  return fkiss_functions[fno].funcname;
}

/***************************************************************************/

static int match_function_name(char *name,int fno)
{

  while ( fkiss_functions[fno].funcname &&
            (strcmp(fkiss_functions[fno].funcname,name) != 0) )
      fno++;

  return fno;
}

/***************************************************************************/

BOOL IsFKISSFunction(char *s)
{
  char funcname[MAX_FUNCNAME_LEN+1], *p;
  int len=0, fno;

  skipspace(&s);
  p = s;

  while ( isalpha(*p) )
  {
    p++;
    len++;
  }

  strncpy(funcname,s,len);
  funcname[len] = 0;

  if ( *p != '(' )
    return FALSE;

  lowercase(funcname,funcname,MAX_FUNCNAME_LEN+1);

  fno = match_function_name(funcname,0);

  return (BOOL)(fkiss_functions[fno].funcname != NULL );
}

/***************************************************************************/

static int findsamefid(int fno)
{
  int fother = 0, fid = fkiss_functions[fno].fid;

  while ( (fkiss_functions[fother].fid != fid) || (fother == fno) )
    fother++;

  return fother;
}

/***************************************************************************/

#define PARAMBLOCKSIZE (MAXPARAMCOUNT * sizeof(FKISS_PARAMETER))

#define ERROR(id) { *fkiss = q; return id; }

static int fkiss_command(char **fkiss,BOOL *skiptoend)
{
  char *params[MAXPARAMCOUNT], funcname[MAX_FUNCNAME_LEN+1];
  char paramblock[PARAMBLOCKSIZE];
  char *q, c, *p;
  FKISS_PARAMETER paramval;
  int pcount, fno=0, ptype, pno, len, loop, blocksize, paramsize, preqd;
  int status, errortp, lastfno;
  FKISS_ACTION_DETAILS action;
  FKISS_EVENT_DETAILS event;

  q = *fkiss;
  *skiptoend = TRUE;

  /* Extract the function name */

  while ( isalpha(*q) ) q++;
  if ( !*q )
    return FKS_P_MissFunc;

  if ( *q != '(' )
    ERROR(FKS_P_BadSyntax)

  *q = 0;

  strcpy(funcname,*fkiss);

  p = strchr(funcname,' ');
  if ( p ) *p = 0;

  /* Next get all the parameters */

  for ( pcount = 0; pcount < MAXPARAMCOUNT; pcount++ )
    params[pcount] = NULL;

  memset(&action,0,sizeof(FKISS_ACTION_DETAILS));
  memset(&event,0,sizeof(FKISS_EVENT_DETAILS));
  memset(paramblock,0,PARAMBLOCKSIZE);

  pcount = 0;

  /*
   * Perform full parse operation on fkiss parameters
   */

  q++;
  skipspace(&q);
  c = *q;

  while ( c && (c != ')') )
  {
    skipspace(&q);

    params[pcount++] = q;
    switch (*q)
    {
      case '!' :
           q++;
           if ( !isalpha(*q) )
             ERROR(FKS_P_BadSyntax);

      default  :
          if ( isalpha(*q) )
          {
            while (isalnum(*q) ) q++;
            break;
          }
          if ( !isdigit(*q) )
            ERROR(FKS_P_BadSyntax)

      case '-' :
      case '#' :
          q++;
          while ( isdigit(*q) ) q++;
          break;

      case '\"' :
          q++;
          while ( *q && (*q != '\"') )
          {
            if (*q == '\\' ) q++;
            q++;
          }
          q++;
          break;

      case ' ' : break;
      case ',' :
      case ')' : ERROR(FKS_P_BadSyntax)
    }

    /* Have parsed a single parameter */

    c  = *q;
    *q = 0;
    q++;
    if ( isspace(c) )
    {
      skipspace(&q);
      c = *q;
      q++;
    }
    if ( (c != ',') && (c != ')') )
      ERROR(FKS_P_BadSyntax)
  }

  *skiptoend = FALSE;

  if ( pcount == 0 ) q++;
  *fkiss = q;

  /* Ensure there is a function with the correct name */

  fno = match_function_name(funcname,0);
  if ( fkiss_functions[fno].funcname == NULL )
    return FKS_P_BadFunc;

  errortp = FKISS_ERROR(fkiss_functions[fno]);
  switch ( errortp )
  {
    case FKISS_MISSPELLING :
            message_error("FKS29",
                         fkiss_functions[findsamefid(fno)].funcname);
            break;

    case FKISS_SYNONYM :
            message_error("FKS30",
                         fkiss_functions[fno].funcname,
                         fkiss_functions[findsamefid(fno)].funcname);
            break;

    case FKISS_UNSUPPORTED :
            message_error("FKS31",fkiss_functions[fno].funcname);
            break;

    case FKISS_NOTPORTABLE :
            message_error("FKS32",fkiss_functions[fno].funcname);
            break;

  }

  /*
   * Syntactic check for close bracket
   */

  if ( !c )
    return FKS_P_BadSyntax;  /* No close bracket */

  /*
   * Basic parse completed
   * Now validate the required function parameters
   * against the parameters actually provided for each of
   * the functions with the correct name
   */

  while ( fkiss_functions[fno].funcname != NULL )
  {
    blocksize = 0;
    lastfno   = fno;
    status    = 0;

    for ( preqd = 0; preqd < MAXPARAMCOUNT; preqd++ )
    {
      ptype = PARAMETERTYPE(fkiss_functions[fno].params,preqd);
      if ( ptype == 0 ) break;
    }

    if ( pcount == preqd )
    {
      for ( pno = 0; pno < MAXPARAMCOUNT; pno++ )
      {
        ptype = PARAMETERTYPE(fkiss_functions[fno].params,pno);
        if ( ptype == 0 ) break;

        switch ( ptype )
        {
          case fks_p_celgrp :
                 status = read_celgrp(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_object :
                 status = read_object(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_objorvar :
                 status = read_objorvar(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_celgrpobj :
                 status = read_cog(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_grp :
                 status = read_group(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_cogv :
                 status = read_cogv(params[pno],&(paramval.instance));
                 paramsize = sizeof(FKISS_INSTANCE_TYPE);
                 break;

          case fks_p_string :
                 paramval.string = read_string(params[pno]);
                 status = (paramval.string == NULL)? FKS_P_BadString : 0;
                 paramsize = sizeof(char *);
                 break;

          case fks_p_file :
                 paramval.filename = read_string(params[pno]);
                 status = (paramval.filename == NULL)? FKS_P_BadFile : 0;
                 if ( !status )
                 {
                   if ( strlen(paramval.filename) > MAXDOSNAMELEN )
                     status = FKS_P_FileTooLong;
                 }
                 paramsize = sizeof(char *);
                 if ( !status )
                 { len = strlen(paramval.filename);
                   for ( loop = 0; (loop < len) && !status; loop++)
                     if ( !isgraph(paramval.filename[loop]) ||
                           isspace(paramval.filename[loop]) )
                       status = FKS_P_BadFile;
                   if ( !status )
                   {
                     if ( len > 0 )
                     {
                       if ( len < 5 ) status = FKS_P_BadFile;
                       else
                       {
                         q = strchr(paramval.filename,'.');
                         if ( !q || (strlen(q)>4) )
                           status = FKS_P_BadFile;
                       }
                     }
                   }
                 }
                 break;

          case fks_p_constant :
                 paramval.intval.flags = 0;
                 if ( !convert_to_int(params[pno],&(paramval.intval.val.constant)) )
                   status = FKS_P_BadInt;
                 paramsize = sizeof(FKISS_INTEGERVALUE);
                 break;

          case fks_p_variable :
                 paramval.variable = read_variable(params[pno],VARIABLE_K_VALUE);
                 status = (paramval.variable == NULL) ? FKS_P_BadVar : 0;
                 paramsize = sizeof(FKISS_VARIABLE_LIST);
                 break;

          case fks_p_intval :
                 paramsize = sizeof(FKISS_INTEGERVALUE);
                 paramval.intval.flags = FKINT_K_VARIABLE;
                 paramval.intval.val.variable = read_variable(params[pno],0);
                 if ( paramval.intval.val.variable == NULL )
                 { /* A literal integer is also legal */
                   paramval.intval.flags = 0;
                   if ( !convert_to_int(params[pno],&(paramval.intval.val.constant)) )
                   {
                     status = FKS_P_BadInt;
                     break;
                   }
                 }
                 status = 0;
                 break;

          case fks_p_keyid :
                 paramval.keyid = read_keyid(params[pno]);
                 status = (paramval.keyid < 0) ? FKS_P_BadKeyName : 0;
                 paramsize = sizeof(char *);
                 break;

          case fks_p_idtag :
                 paramval.idtag = read_string(params[pno]);

                 /*
                  * We assume here that only fks_valuepool uses an idtag parameter
                  * but disallows the <pool>.<tag> format.
                  */
                 status = 0;

                 if ( ! poolio_validateID(paramval.idtag,
                                          fkiss_functions[fno].fid != fks_valuepool) )
                   status = FKS_P_BadTag;
                 paramsize = sizeof(char *);
                 break;

          default : fkiss_error("FKP2","fkiss_command");
        }  /* Switch */

        if ( status )
          break;

        memcpy(&(paramblock[blocksize]),&paramval,paramsize);
        blocksize += paramsize;

      } /* for */
    }   /* if */
    else status = FKS_P_BadParams;

    if ( !status )
      break;

    fno = match_function_name(funcname,fno+1);

  } /* while */

  if ( status )
  {
    if ( fkiss_functions[fno].eventfn )
      skip_to_next_event = TRUE;
    return status;
  }

  KISS_Version = MAX(KISS_Version,FKISS_VERSION(fkiss_functions[fno]));

  /* We now know that the parameters are legal, so either call the
     relevant event routine, or set up the action */

  if ( fkiss_functions[fno].eventfn )
  {
    assert(blocksize <= sizeof(FKISS_EVENT_DETAILS));
    skip_to_next_event = FALSE;
    memcpy(&event,paramblock, blocksize);
    return fkiss_functions[fno].eventfn(fkiss_functions[fno].fid,&event);
  }
  else
    if ( !skip_to_next_event )
    {
      assert(blocksize <= sizeof(FKISS_ACTION_DETAILS));
      memcpy(&action,paramblock,blocksize);
      return fkiss_add_action(fkiss_functions[fno].fid,&action);
    }

  return FKS_P_Success;
}

/***************************************************************************/

void process_fkiss_command_line(char *fkiss)
{
  int status;
  BOOL skiptoend;

  skipspace(&fkiss);

  while ( *fkiss && (*fkiss != ';') )
  {
    save_fkiss_text(&fkiss);

    status = fkiss_command(&fkiss,&skiptoend);
    if ( status != FKS_P_Success )
    {
      fkiss_report(status);

      if ( skiptoend && *fkiss && (*fkiss != ')') )
      {
        do fkiss++;
        while ( *fkiss && (*fkiss != ')') && (*fkiss != ' ') );
      }
      if ( *fkiss == ')' ) fkiss++;
    }

    skipspace(&fkiss);
  }

  fkiss_flags |= FKISS_K_Active;
}

/***************************************************************************/

typedef BOOL (*action_list_function)(FKISS_ACTION_LIST *action,void *handle);
typedef BOOL (*event_list_function)(FKISS_EVENT_LIST *event,void *handle);

/***************************************************************************/
/*
 * This routine works as follows:
 *   A list of events is provided, normally one of the 5 master lists.
 *   Two functions are provided, one applying to actions and one to events.
 *   The flags word specifies whether the function is called for every action
 *   or just the first in any one list, ditto for events.
 *   The handle is a parameter to the provided functions
 */

#define FKISS_APPLY_ALLEVENTS  (1u<<0)
#define FKISS_APPLY_ALLACTIONS (1u<<1)

static BOOL apply_to_event_list(int                  flags,
                                FKISS_EVENT_LIST    *event,
                                event_list_function  eventfunc,
                                action_list_function actionfunc,
                                void *handle)
{
  FKISS_EVENT_LIST evcpy = *event;
  FKISS_ACTION_LIST action;

  if ( (flags & FKISS_APPLY_ALLEVENTS) || actionfunc )
  {
    while (evcpy)
    {
      if ( (flags & FKISS_APPLY_ALLEVENTS) && eventfunc )
      {
        if ( !eventfunc(&evcpy,handle) )
          return FALSE;
      }

      if ( actionfunc )
      {
        if ( !(flags & FKISS_APPLY_ALLACTIONS) )
        {
          if ( !actionfunc(&evcpy->actions,handle) )
            return FALSE;
        }
        else
        {
          action = evcpy->actions;
          while ( action )
          {
            if ( !actionfunc(&action,handle) )
              return FALSE;
            action = action->next_action;
          }
        }
      }
      evcpy = evcpy->next_event;
    }
  }

  if ( !(flags & FKISS_APPLY_ALLEVENTS) && eventfunc )
    return eventfunc(event,handle);
  return TRUE;
}

/***************************************************************************/

BOOL fkiss_apply_to_all(int flags,
                        event_list_function  eventfunc,
                        action_list_function actionfunc,
                        void *handle)
{  return apply_to_event_list( flags,
                               &fkiss_misc_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_alarm_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_mouse_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_collision_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_label_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_key_list,
                               eventfunc,
                               actionfunc,
                               handle) &&
          apply_to_event_list( flags,
                               &fkiss_pointer_list,
                               eventfunc,
                               actionfunc,
                               handle);
}

/***************************************************************************/

static void raise_unref(BOOL *hdr_given,char *tag,int lineno)
{
  if ( !*hdr_given )
  {
    *hdr_given = TRUE;
    message_error("UNREF");
  }
  message_error(tag,lineno);
}

/***************************************************************************/
/* Returns TRUE if the intergervalue is a constant, /and/ it is equal to
   the sample value.
 */

static BOOL isalwaysequal(FKISS_INTEGERVALUE ival, int sample)
{
  return ((ival.flags & FKINT_K_VARIABLE) == 0) &&
          (ival.val.constant == sample);
}

/***************************************************************************/
/*
 * Must return /FALSE/ if the action references the alarm event.
 */

static BOOL check_unref_timer(FKISS_ACTION_LIST *action,void *e)
{
  FKISS_EVENT_LIST event = e;
  FKISS_INTEGERVALUE alarm = event->data.alarm.alarm;
  FKISS_INTEGERVALUE timer;

  switch( (*action)->fid )
  {
    case fks_timer :
    case fks_randomtimer :
           timer = (*action)->data.timer.alarm;
           break;

    case fks_iffixed :
    case fks_ifnotfixed :
    case fks_ifmapped :
    case fks_ifnotmapped :
    case fks_ifmoved :
    case fks_ifnotmoved :
           timer = (*action)->data.iftimer.alarm;
           break;

    default : return TRUE;
   }

  if ( (alarm.flags == timer.flags) &&
       (alarm.val.constant == timer.val.constant) )
    return FALSE;

  return TRUE;
}

/***************************************************************************/

static BOOL eventMatches(FKISS_EVENT_LIST   event,
                         FKISS_INTEGERVALUE target,
                         FKISS_INTEGERVALUE jumpto,
                         BOOL               islabel)
{
  if ( (target.flags & FKINT_K_TYPE) == FKINT_K_CONSTANT )
    return isalwaysequal(jumpto,target.val.constant);

  /*
   * The target label is an identifier.
   */

  if ( (jumpto.flags & FKINT_K_TYPE) == FKINT_K_CONSTANT )
    return FALSE;

  /*
   * The action label or alarm is an identifier - check it has the correct
   * "label" or "alarm" value.
   */

  return (islabel ? jumpto.val.variable->label : jumpto.val.variable->alarm) == event;
}

#define labelMatches(e,l) eventMatches(e,e->data.label,l,TRUE)
#define alarmMatches(e,t) eventMatches(e,e->data.alarm.alarm,t,FALSE)

/***************************************************************************/
/*
 * Must return /FALSE/ if the action references the label event.
 */

static BOOL check_unref_label(FKISS_ACTION_LIST *action,void *e)
{
  FKISS_EVENT_LIST event = e;

  switch ( (*action)->fid )
  {
    case fks_goto :
    case fks_gosub :
           return !labelMatches(event,(*action)->data.label);

    case fks_gotorandom :
    case fks_gosubrandom :
           return !(labelMatches(event,(*action)->data.branch.label1) ||
                    labelMatches(event,(*action)->data.branch.label2) );

    case fks_repeat :
           return !labelMatches(event,(*action)->data.repeat.label);
  }
  return TRUE;
}

/***************************************************************************/

static void check_unref_events(FKISS_EVENT_LIST event,
                               action_list_function actionfunc,
                               BOOL *hdr_given, char *tag)
{

  while ( event )
  {
    if ( fkiss_apply_to_all(FKISS_APPLY_ALLACTIONS,NULL,actionfunc,event) )
      raise_unref(hdr_given,tag,event->lineno);
    event = event->next_event;
  }
}

/***************************************************************************/

static BOOL check_label_is(FKISS_EVENT_LIST *event,void *handle)
{
  FKISS_INTEGERVALUE *ival = (FKISS_INTEGERVALUE *)handle;
  return !labelMatches((*event),*ival);
}

/***************************************************************************/

static BOOL check_alarm_is(FKISS_EVENT_LIST *event,void *handle)
{
  FKISS_INTEGERVALUE *ival = (FKISS_INTEGERVALUE *)handle;

  return !alarmMatches((*event),*ival);
}

/***************************************************************************/
/*
 * Must return FALSE is the action has a handler
 */

#define ISCONSTANT(x) ((x.flags & FKINT_K_VARIABLE) == 0)

#define ISHANDLER(intval,list,func) \
      ISCONSTANT(intval) \
       ? apply_to_event_list(FKISS_APPLY_ALLEVENTS,&list,func,NULL,&intval) \
       : FALSE

static BOOL check_unhandled_actions(FKISS_ACTION_LIST *action,void *handle)
{
  BOOL *hdr_given = (BOOL *)handle;
  BOOL no_handler = TRUE;
  int tagid;

  switch ( (*action)->fid )
  {
    case fks_timer :
    case fks_randomtimer :
             no_handler = ISHANDLER((*action)->data.timer.alarm,
                                    fkiss_alarm_list,
                                    check_alarm_is);
             tagid = 1;
             break;

    case fks_iffixed :
    case fks_ifnotfixed :
    case fks_ifmapped :
    case fks_ifnotmapped :
    case fks_ifmoved :
    case fks_ifnotmoved :
             no_handler = ISHANDLER((*action)->data.iftimer.alarm,
                                    fkiss_alarm_list,
                                    check_alarm_is);
             tagid = 1;
             break;

    case fks_goto :
    case fks_gosub :
             no_handler = ISHANDLER((*action)->data.label,
                                    fkiss_label_list,
                                    check_label_is);
             tagid = 2;
             break;

    case fks_gotorandom :
    case fks_gosubrandom :
             no_handler = ISHANDLER((*action)->data.branch.label1,
                                    fkiss_label_list,
                                    check_label_is) ||
                          ISHANDLER((*action)->data.branch.label2,
                                    fkiss_label_list,
                                    check_label_is);
             tagid = 2;
             break;

    default : return TRUE;
  }

  if ( no_handler )
     raise_unref(hdr_given,tagid == 1 ? "UNREF03" : "UNREF04" ,(*action)->lineno);

  return TRUE;
}

/***************************************************************************/

BOOL fkiss_load_completed(void)
{
  FKISS_EVENT_LIST    event;
  FKISS_VARIABLE_LIST var;
  BOOL hdr_given;
  char varname[MaxIdentifierLen+1];

  /* Check that any logical if blocks are closed */

  if ( LogicalIfState != IfState_None )
    message_error("FKS19a",begin_logical_if_line);

  LogicalIfState = IfState_None;

  /*
   * Make variables associated with symbolically-named labels point to
   * the label event.
   */

  event = fkiss_label_list;
  while ( event )
  {
    if ( (event->data.label.flags & FKINT_K_TYPE) == FKINT_K_VARIABLE )
    {
      var        = event->data.label.val.variable;
      var->label = event;
    }
    event = event->next_event;
  }

  /*
   * Ditto for symbolically-named alarms
   */

  event = fkiss_alarm_list;
  while ( event )
  {
    if ( (event->data.alarm.alarm.flags & FKINT_K_TYPE) == FKINT_K_VARIABLE )
    {
      var        = event->data.alarm.alarm.val.variable;
      var->alarm = event;
    }
    event = event->next_event;
  }


  /* Check for any unresolved references */

  hdr_given = FALSE;

  check_unref_events(fkiss_alarm_list,check_unref_timer,&hdr_given,"UNREF01");
  check_unref_events(fkiss_label_list,check_unref_label,&hdr_given,"UNREF02");
  fkiss_apply_to_all(FKISS_APPLY_ALLACTIONS,NULL,check_unhandled_actions,&hdr_given);

  var = fkiss_variables;
  while ( var )
  {
    if ( ((var->flags & VARIABLE_K_TYPEMASK) != VARIABLE_K_VALUE) &&
         ((var->flags & VARIABLE_K_TYPEMASK) != VARIABLE_K_LABEL) &&
         ((var->flags & VARIABLE_K_TYPEMASK) != VARIABLE_K_ALARM) )
    {
      strncpy(varname,var->name,MaxIdentifierLen);
      varname[MaxIdentifierLen] = 0;
      message_error((var->flags == 0) ? "BADVAR0" : "BADVAR1" ,varname);
    }
    var = var->next;
  }

  return TRUE;
}

/***************************************************************************/

static BOOL clear_fkiss_actions(FKISS_ACTION_LIST *list,void *handle)
{
  FKISS_ACTION_LIST curr=*list,last;

  *list = NULL;

  while ( curr )
  {
    last = curr;
    curr = curr->next_action;
    switch ( last->fid )
    {
      case fks_notify :
      case fks_debug  : free(last->data.debugmsg);
                        break;
      case fks_sound  :
      case fks_music  : free(last->data.sound);
                        break;

      case fks_letkey :
      case fks_letkeymap : free(last->data.key.keylist);
                           break;
    }
    free(last);
  }
  return TRUE;
}

/***************************************************************************/

static BOOL clear_fkiss_events(FKISS_EVENT_LIST *event,void *handle)
{
  FKISS_EVENT_LIST temp;

  while (*event)
  {
    temp   = *event;
    *event = temp->next_event;
    free(temp);
  }
  return TRUE;
}

/***************************************************************************/

void fkiss_clear_doll(void)
{
  FKISS_VARIABLE_LIST v;

  fkiss_apply_to_all(0,clear_fkiss_events,clear_fkiss_actions,NULL);

  currentevent = NULL;
  skip_to_next_event = FALSE;

  while ( fkiss_variables )
  {
    v               = fkiss_variables;
    fkiss_variables = fkiss_variables->next;
    free(v);
  }
}
