/***************************************************************************/
/*                                                                         */
/*                    Front end for KISS doll validator                    */
/*                                                                         */
/***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "BOOL.h"

#include "validate.h"
#include "utilities.h"
#include "KISS_data.h"
#include "message.h"
#include "celgroup.h"
#include "osspecific.h"
#include "lzh.h"

#define QUAL_SYNTAXONLY 0
#define QUAL_DOLLSTATS  1
#define QUAL_OUTPUT     2
#define QUAL_FKISSCASE  3
#define QUAL_ALLCNF     4
#define QUAL_PROGRESS   5
#define QUAL_TEMPDIR    6

typedef struct _qual {
                       char *qual;
                       char *helptxt;
                     } QUALIFIER_TYPE;

static const QUALIFIER_TYPE qualifiers[] =
  {
    { "syntax_only",       "Suppress cel & kcf checks" },
    { "report",            "Generate brief statistics on doll" },
    { "output <filename>", "Send output to <filename>" },
    { "case_insensitive",  "Suppress case checks on FKISS function names" },
    { "all",               "Check all cnf files found (LZH mode only)" },
    { "progress",          "Give progress indicator" },
    { "temp_dir",          "Specify alternate temporary directory" },
    {  NULL, NULL }
  };

static const char *VersionString = "0.14 August 2011";

/***************************************************************************/

static void give_help(void)
{
  int qidx=0;

  printf("\ncnfcheck %s\n",VersionString);

  printf("\nFormat:\n   cnfcheck [-options] [<cnffile>|<lzhfile>]\n");
  printf("\nOptions are:\n");
  while ( qualifiers[qidx].qual )
  {
    printf("   -%-20s %s\n",qualifiers[qidx].qual,qualifiers[qidx].helptxt);
    qidx++;
  }

}

/***************************************************************************/

static int getQualifierIdx(char *qual)
{
  int qidx=0, len = strlen(qual)-1;

  if ( len > 0 )
  {
    qual++;

    while ( qualifiers[qidx].qual )
    {
      if ( struncmp(qualifiers[qidx].qual,qual,len) == 0 )
        return qidx;
      qidx++;
    }
  }
  return -1;
}

/***************************************************************************/

static void report_doll_statistics(char *cnf)
{

  message_reportTag("STAT0",cnf);
  message_reportTag("STAT1",CelCount);
  message_reportTag("STAT2",PaletteCount);
  message_reportTag("STAT3",ObjectCount);
  if ( (ActiveSetsMask & 0x3ff) == 0x3ff )
    message_reportTag("STAT4");
  else
  {
    int setno;
    message_reportTag("STAT5");
    for ( setno = 0; setno < MAX_SETS; setno++ )
      if ( ActiveSetsMask & (1u<<setno) )
        message_report(" %1d",setno);
    message_report("\n");
  }
  if ( !syntax_only )
  {
    char tagbuff[16];

    sprintf(tagbuff,"KVER%1d",KISS_Version);
    message_reportTag("KVER",message_msgtext(tagbuff));

    sprintf(tagbuff,"KCM%1d",KISS_ColourModel);
    message_reportTag("KCM",message_msgtext(tagbuff));
  }

}

/***************************************************************************/

static void processCNF(char *cnfname,BOOL report)
{
  message_progress(message_msgtext("ReadCNF"),cnfname);
  if ( KISS_validateDoll(cnfname) )
  {
    if (errorcount == 0)
      message_reportTag("Success",cnfname);

    if ( report )
      report_doll_statistics(cnfname);
  }
}

/***************************************************************************/

int main(int argc,char **argv)
{
  int qidx=1, status;
  BOOL report_stats=FALSE,allcnf=FALSE;
  char *outname = NULL, *inname, *p;
  LZH_CONTENT_LIST cnffile;

  /*
   * Firstly, make sense of any qualifiers on the command line.
   */

  while ( (qidx < (argc-1)) && (*(argv[qidx]) == '-') )
  {
    switch ( getQualifierIdx(argv[qidx]) )
    {
      case QUAL_SYNTAXONLY :
            syntax_only = TRUE;
            break;

      case QUAL_DOLLSTATS :
            report_stats = TRUE;
            break;

      case QUAL_OUTPUT :
            qidx++;
            outname = argv[qidx];
            break;

      case QUAL_FKISSCASE :
            case_sensitive_fkiss = FALSE;
            break;

      case QUAL_ALLCNF :
            allcnf = TRUE;
            break;

      case QUAL_PROGRESS :
            show_progress = TRUE;
            break;

      case QUAL_TEMPDIR :
            qidx++;
            working_directory = argv[qidx];
            break;

      default :
            printf("Unknown qualifier \'%s\'\n",argv[qidx]);
            give_help();
            exit(1);
            break;
    }
    qidx++;
  }

  /*
   * Is there still a filename given ?
   */

  if ( qidx >= argc )
  {
    give_help();
    exit(0);
  }

  message_init(outname);

  inname = argv[qidx];

  /*
   * Deduce whether this is a cnf file or an lzh archive
   */

  p = strrchr(inname,FILEEXT);
  if ( p && (strucmp(++p,"cnf") == 0) )
  {
    processCNF(inname,report_stats);
    return 0;
  }

  /*
   * Not a cnf - presumably an lzh archive
   */

  /*
   * Firstly, we need to create a working directory
   */

  if ( !working_directory )
  {
    time_t t;
    char buff[FILENAME_MAX];

    t = time(NULL);
    sprintf(buff,"%08x",t);
    working_directory        = strdup(buff);
    remove_working_directory = TRUE;
    lzh_processCommand(CREATEDIR,NULL,NULL);
  }

  lzhArchiveName = inname;
  status = lzh_scanFile(lzhArchiveName,&lzhArchiveContents);

  atexit(lzh_deleteTemporaryFiles);

  switch ( status )
  {
    case LZH_NOTFOUND :
      fprintf(stderr,"File %s not found\n",lzhArchiveName);
      exit(1);

    case LZH_INVALID :
      fprintf(stderr,"Invalid or corrupt lzh file %s\n",lzhArchiveName);
      exit(1);
  }

  cnffile = lzh_findCNF(lzhArchiveContents);
  while ( cnffile )
  {
    processCNF(cnffile->filename,report_stats);
    if ( !allcnf )
      break;
    cnffile = lzh_findCNF(cnffile->next);
  }
  return 0;
}
