/*----------------------------------------------------------------------*/
/*                                                                      */
/*                         Mini LZH handler                             */
/*                                                                      */
/*         Depends on standard 'lharc' command being in CLI path        */
/*                                                                      */
/*----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#include "BOOL.h"
#include "utilities.h"
#include "osspecific.h"
#include "message.h"
#include "KISS_data.h"

#define EXECUTE(cmnd) system(cmnd)

static LZH_CONTENT_LIST files_to_delete = NULL;

/*----------------------------------------------------------------------*/

static BOOL read_contents(FILE *fp,char *buff)
{
  int ch;

  ch = fgetc(fp);
  if ( ch == EOF )
    return FALSE;

  while ( (ch != EOF) && (ch != '\n') )
  {
    *buff = ch;
    buff++;
    ch = fgetc(fp);
  }
  *buff = 0;
  return TRUE;
}

/*----------------------------------------------------------------------*/

static int findNameColumn(FILE *afp,char *buff)
{
  char *nameptr;
  int nameLen = -1;

  while ( read_contents(afp,buff) )
  {
    nameptr = strstr(buff,"NAME");
    if ( !nameptr )
      nameptr = strstr(buff,"Name");
    if ( nameptr )
    {
      if ( !(nameptr[4]) || isspace(nameptr[4]) )
      {
        nameLen = (int)nameptr - (int)buff;
        break;
      }
    }
  }

  /*
   * Have we come to the end of the file?
   */

  if ( nameLen < 0 )
    return nameLen;

  /*
   * Now move the nameptr back to the start of the underlining on the next line
   */

  if ( !read_contents(afp,buff) || (strlen(buff)<nameLen) )
    return -1;

  nameptr = &(buff[nameLen]);
  while ( (nameptr != buff) && (!isspace(nameptr[-1])) ) nameptr--;
  return (int)nameptr - (int)buff;
}

/*----------------------------------------------------------------------*/

static char *getFileName(FILE *afp,char *buff,int namepos)
{
  char *filename,*p;

  if ( !read_contents(afp,buff) || (strlen(buff)<namepos) )
    return NULL;

  filename = &(buff[namepos]);
  while ( (filename != buff) && !isspace(filename[-1]) )
    filename--;

  while (isspace(*filename)) filename++;

  p = filename;
  while ( *p && !isspace(*p) ) p++;
  *p = 0;

  /*
   * Standardize file extension character to unix type
   */

  if ( FILEEXT != '.' )
  {
    p = strchr(filename,FILEEXT);
    if ( p )
      *p = '.';
  }

  return filename;
}

/*----------------------------------------------------------------------*/

static int tidy(FILE *fp,char *filename,int status)
{
  fclose(fp);
  remove(filename);
  return status;
}

/*----------------------------------------------------------------------*/

void lzh_processCommand(const char *format,
                        const char *archiveName,
                        const char *filename)
{
  char cmnd[256], *p;

  p = cmnd;

  while ( *format )
  {
    if ( *format == '%' )
    {
      format++;
      switch ( *format )
      {
        case 'a' :
          strcpy(p,archiveName);
          p = strchr(p,0);
          break;

        case 'd' :
          strcpy(p,working_directory);
          p = strchr(p,0);
          break;

        case 'f' :
          strcpy(p,filename);
          p = strchr(p,0);
          break;

        default :
          message_report("Internal error : unexpected formatting character\n");
          exit(1);
      }
      format++;
    }
    else
    {
      *p = *format;
      format++;
      p++;
    }
  }

  *p = 0;

  EXECUTE(cmnd);
}

/*----------------------------------------------------------------------*/

void lzh_deleteTemporaryFiles(void)
{
  LZH_CONTENT_LIST tmp;

  while ( files_to_delete )
  {
    tmp             = files_to_delete;
    files_to_delete = files_to_delete->next;
    remove(tmp->filename);
    free(tmp->filename);
    free(tmp);
  }

  if ( remove_working_directory )
    lzh_processCommand(DELETEDIR,NULL,NULL);
}

/*----------------------------------------------------------------------*/

static void addToDeleteList(char *filename)
{
  LZH_CONTENT_LIST thisFile;

  thisFile           = safe_alloc(sizeof(LZH_CONTENT_TYPE));
  thisFile->filename = strdup(filename);
  thisFile->next     = files_to_delete;
  files_to_delete    = thisFile;
}

/*----------------------------------------------------------------------*/

int lzh_scanFile(const char *archivename,LZH_CONTENT_LIST *contents)
{
  LZH_CONTENT_LIST lastFile = NULL, thisFile;
  FILE *afp;
  char buff[256], tmpfile[256], *p, *filename;
  int namePos;

  afp = fopen(archivename,"rb");
  if ( !afp )
    return LZH_NOTFOUND;

  fclose(afp);

  message_progress(message_msgtext("ScanLZH"));

  sprintf(tmpfile,"%s%c%s",working_directory,DIRSEP,archivename);
  p = strrchr(tmpfile,FILEEXT);
  if ( !p )
  {
    p = strchr(tmpfile,0);
    *p = FILEEXT;
  }
  p++;
  strcpy(p,"lis");

  lzh_processCommand(LZHLIST,archivename,tmpfile);

  *contents = NULL;

  afp       = fopen(tmpfile,"r");
  if ( !afp )
    return LZH_INVALID;

  /*
   * Find the header line
   */

  if ( namePos = findNameColumn(afp,buff), namePos < 0 )
    return tidy(afp,tmpfile,LZH_INVALID);

  /*
   * Now read the remainder of the contents list until we find the 'totals'
   * line (indicated by a filename consisting exclusively of '-' characters).
   */

  do
  {
    filename = getFileName(afp,buff,namePos);
    if ( !filename )
      return tidy(afp,tmpfile,LZH_INVALID);

    /*
     * Have we reached the totals line?
     */

    if ( strspn(filename,"-") == strlen(filename) )
      break;

    /*
     * Not the totals line - add it to the contents list.
     */

    thisFile           = safe_alloc(sizeof(LZH_CONTENT_TYPE));
    thisFile->filename = strdup(filename);
    thisFile->next     = NULL;
    if ( lastFile )
      lastFile->next = thisFile;
    else *contents = thisFile;
    lastFile = thisFile;

  } while (TRUE);

  fclose(afp);
  remove(tmpfile);

  /*
   * If it's not syntax-only, extract everything at once - more
   * efficient than extracting single files at a time later.
   */

  if ( !syntax_only )
  {
    message_progress(message_msgtext("MeltLZH"));
    lzh_processCommand(LZHEXTRACTALL,archivename,"");

    /*
     * Files have been extracted - add the entire contents to the list of files
     * that need to be deleted.
     */

    thisFile = *contents;
    while ( thisFile )
    {
      sprintf(tmpfile,"%s%c%s",working_directory,DIRSEP,thisFile->filename);
      if ( FILEEXT != '.' )
      {
        p = strrchr(tmpfile,DIRSEP);
        if ( p )
        {
          p = strchr(p,'.');
          if ( p ) *p = FILEEXT;
        }
      }
      addToDeleteList(tmpfile);
      thisFile = thisFile->next;
    }
  }

  return LZH_OK;
}

/*----------------------------------------------------------------------*/

LZH_CONTENT_LIST lzh_findCNF(LZH_CONTENT_LIST contents)
{
  char *p;

  while ( contents )
  {
    p = strrchr(contents->filename,'.');
    if ( p )
    {
      p++;
      if ( strucmp(p,"cnf") == 0 )
        break;
    }
    contents = contents->next;
  }
  return contents;
}

/*----------------------------------------------------------------------*/

LZH_CONTENT_LIST lzh_findFile(LZH_CONTENT_LIST contents,const char *filename)
{
  while ( contents )
  {
    if ( strucmp(contents->filename,filename) == 0 )
      break;
    contents = contents->next;
  }
  return contents;
}

/*----------------------------------------------------------------------*/

FILE *lzh_openFile(const char       *archiveName,
                   LZH_CONTENT_LIST contents,
                   const char       *filename,
                   const char       *mode)
{
  char fullname[256], adjustedName[256], *p, *q, *usename;
  LZH_CONTENT_LIST filePtr;
  FILE *fp;

  if ( filePtr = lzh_findFile(contents,filename), filePtr == NULL )
    return NULL;

  /*
   * Allow for non-Unix file extension character
   */

  if ( FILEEXT != '.' )
  {
    p = filePtr->filename;
    q = adjustedName;

    while ( *p )
    {
      if ( *p == '.' )
        *q = FILEEXT;
      else *q = *p;
      p++;
      q++;
    }
    *q = 0;
    usename = adjustedName;
  }
  else usename = filePtr->filename;

  /*
   * What is the full filename of the extracted file?
   */

  sprintf(fullname,"%s%c%s",working_directory,DIRSEP,usename);

  if ( syntax_only )  /* We haven't already extracted the entire archive */
  {
    lzh_processCommand(LZHEXTRACT,archiveName,usename);
    addToDeleteList(fullname);
  }

  fp = fopen(fullname,mode);

  return fp;
}
