/*======================================================================================*/
/*                                                                                      */
/*                          KISS Pragma handler                                         */
/*                                                                                      */
/*======================================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "BOOL.h"
#include "KISS_data.h"
#include "readkiss.h"
#include "message.h"
#include "utilities.h"

#define PRAGMA_UNSET ((int)0x80000000)

typedef struct {
                 char *key;
                 int  keyValue;
               } PRAGMA_KEY_TYPE;

typedef struct {
                 char            *pragmaName;
                 unsigned int     valueType;
                 int              value;
                 COORD_PAIR       coords;
                 PRAGMA_KEY_TYPE *keyValue;
               } PRAGMA_TYPE;

static PRAGMA_KEY_TYPE boundsValues[] =
  {
    { "All",     BOUNDS(BOUNDS_K_Full,BOUNDS_K_Full) },
    { "FKISS",   BOUNDS(BOUNDS_K_Off,BOUNDS_K_Full) },
    { "Visible", BOUNDS(BOUNDS_K_Visual,BOUNDS_K_Visual) },
    { "Limited", BOUNDS(BOUNDS_K_Off,BOUNDS_K_Visual) },
    { "None",    BOUNDS(BOUNDS_K_Off,BOUNDS_K_Off) },
    { NULL,      PRAGMA_UNSET }
  };

#define ValueType_K_Number  1
#define ValueType_K_Keyword 2
#define ValueType_K_Coords  3

static PRAGMA_TYPE pragmae[] =
 {
  { "stack",      ValueType_K_Number,  PRAGMA_UNSET, {0,0}, NULL },
  { "maxfix",     ValueType_K_Number,  PRAGMA_UNSET, {0,0}, NULL },
  { "bounds",     ValueType_K_Keyword, PRAGMA_UNSET, {0,0}, boundsValues },
  { "resolution", ValueType_K_Coords,  0,            {PRAGMA_UNSET,PRAGMA_UNSET}, NULL },
  { NULL, 0 }
 };

typedef int (*pragmafn)(PRAGMA_TYPE *pragma,void *handle);

static const char *pragmaKey = "HINT";

/*======================================================================================*/

static BOOL pragmaSet(int pidx)
{
  int pval;
  pval = (pragmae[pidx].valueType == ValueType_K_Coords) ? pragmae[pidx].coords.x
                                                         : pragmae[pidx].value;
  return (pval != PRAGMA_UNSET);
}

/*======================================================================================*/

static int lookupPragma(const char *s)
{
  int idx = 0, len;

  while ( pragmae[idx].pragmaName )
  {
    len = strlen(pragmae[idx].pragmaName);
    if ( struncmp(pragmae[idx].pragmaName,s,len) == 0 )
    {
      if ( isspace(s[len]) )
        return idx;
    }
    idx++;
  }
  return -1;
}

/*======================================================================================*/

static int getKeyValue(PRAGMA_KEY_TYPE *keys, char *key)
{
  int idx=0, len =0;
  char *p=key;

  while ( isalpha(*p) )
  {
    p++;
    len++;
  }

  if ( len > 0 )
  {
    while ( keys[idx].key )
    {
      if ( len <= strlen(keys[idx].key) )
      {
        if ( struncmp(keys[idx].key,key,len) == 0 )
          return idx;
      }
      idx++;
    }
  }
  return PRAGMA_UNSET;
}

/*======================================================================================*/

static void getCoords(char *s,COORD_PAIR *coords)
{
  char *scpy = s;
  int x,y;

  if ( !read_integer(&s,",",&x) )
  {
    message_error("CNF38",scpy);
    return;
  }
  s++;
  skipspace(&s);
  if ( !convert_to_int(s,&y) )
  {
    message_error("CNF38",scpy);
    return;
  }
  coords->x = x;
  coords->y = y;
}

/*======================================================================================*/

static void badPragmaValueType(int pidx)
{
  fprintf(stderr,"Unexpected pragma value type %1d (%s)",
          pragmae[pidx].valueType,
          pragmae[pidx].pragmaName);
  exit(1);
}

/*======================================================================================*/

BOOL pragma_setPragma(char *s)
{
  int pidx, len = strlen(pragmaKey);

  skipspace(&s);

  if ( strncmp(s,pragmaKey,len) != 0 )
    return FALSE;

  s += len;

  skipspace(&s);

  if ( pidx = lookupPragma(s), pidx < 0 )
  {
    message_error("CNF39",s);
    return FALSE;
  }

  s += strlen(pragmae[pidx].pragmaName);

  skipspace(&s);

  switch ( pragmae[pidx].valueType )
  {
    case ValueType_K_Number :
      if ( !isdigit(*s) )
        return FALSE;
      if ( !pragmaSet(pidx) )
        pragmae[pidx].value = atoi(s);
      break;

    case ValueType_K_Keyword :
      if ( !pragmaSet(pidx) )
      {
        pragmae[pidx].value = getKeyValue(pragmae[pidx].keyValue,s);
        if ( pragmae[pidx].value == PRAGMA_UNSET )
          message_error("CNF40",pragmae[pidx].pragmaName,s);
      }
      break;

    case ValueType_K_Coords :
      if ( !pragmaSet(pidx) )
        getCoords(s,&(pragmae[pidx].coords));
      break;

    default : badPragmaValueType(pidx);
  }

  return TRUE;
}

/*======================================================================================*/

void pragma_reset(void)
{
  int pidx=0;

  while ( pragmae[pidx].pragmaName )
  {
    pragmae[pidx].value      = PRAGMA_UNSET;
    pragmae[pidx].coords.x   = PRAGMA_UNSET;
    pragmae[pidx].coords.y   = PRAGMA_UNSET;
    pidx++;
  }
}
