/***************************************************************************/
/*                                                                         */
/*                         Low-level utilities for KISS_IO                 */
/*                                                                         */
/***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <stdarg.h>

#include "BOOL.h"

#include "FileHdrTyp.h"
#include "KISSTypes.h"
#include "FKISSTypes.h"
#include "KISS_data.h"
#include "utilities.h"
#include "message.h"
#include "osspecific.h"
#include "FKISS_IO.h"
#include "pragma.h"
#include "lzh.h"

#define CNF_COMMENT  ';'
#define CNF_FKISS    '@'
#define CNF_MARK     '#'
#define CNF_CELGROUP '!'

static char *pathname = NULL;

static BOOL possible_fkiss_detected = FALSE;

/***************************************************************************/

void skipspace(char **cptr)
{
  while ( isspace(**cptr) ) (*cptr)++;
}

/***************************************************************************/

FILE *fopen_path(char *leafname,char *mode)
{
  char fullname[MaxFileNameLen], leaf[MaxFileNameLen];

  if ( lzhArchiveName )
    return lzh_openFile(lzhArchiveName,lzhArchiveContents,leafname,mode);

  strcpy(fullname,pathname);

  if ( FILEEXT != '.' )
  {
    char leaf[MaxFileNameLen], *p;

    strcpy(leaf,leafname);
    p = strchr(leaf,'.');
    if ( p ) *p = FILEEXT;
    strcat(fullname,leaf);
  }
  else strcat(fullname,lowercase(leafname,leaf,MaxFileNameLen));

  return fopen(fullname,mode);
}

/***************************************************************************/

FILE *fopen_cnf(char *fname,char *mode)
{
  char fullname[MaxFileNameLen],*leafname;
  FILE *fp;

  if ( lzhArchiveName )
    return lzh_openFile(lzhArchiveName,lzhArchiveContents,fname,mode);

  strcpy(fullname,fname);
  leafname = strrchr(fullname,DIRSEP);
  if ( !leafname )
  {
    pathname = "";
    leafname = fullname;
  }
  else
  {
    char *p;

    *leafname = 0;
    pathname  = safe_alloc(strlen(fullname)+2);
    strcpy(pathname,fullname);
    p    = strchr(pathname,0);
    *p   = DIRSEP;
    p[1] = 0;
    leafname++;
  }

  fp = fopen_path(leafname,mode);

  return fp;
}

/***************************************************************************/

BOOL read_next_field(char **cptr, char *termset,char *buffer,
                     int maxbufflen,int *charsread)
{ BOOL success;

  *charsread = 0;
  *buffer    = 0;
  skipspace(cptr);

  while ( (*charsread < maxbufflen ) && **cptr &&
          (strchr(termset,**cptr) == NULL) )
  {
    *buffer = **cptr;
    buffer++;
    *buffer = 0;
    (*charsread)++;
    (*cptr)++;
  }
  success =  (BOOL)( strchr(termset,**cptr) != NULL );
  return success;
}

/***************************************************************************/
#define MAX_INT_LEN 32

BOOL read_integer(char **cptr,char *termset,int *value)
{ char intbuffer[MAX_INT_LEN+1];
  int  len;

  if ( !read_next_field(cptr,termset,intbuffer,MAX_INT_LEN,&len) )
    return FALSE;

  return len ? convert_to_int(intbuffer,value) : FALSE;
}

/***************************************************************************/

int read_line(FILE *fp,char *buffer,int bufflen,
              char *commenttext,int commentlen,
              int *lineno,char **buffstart)
{ int chcount=0, ch = 0, nextch, commcount=0;
  char *chptr,*comptr;
  BOOL comment;

  while ( (chcount == 0) && (ch != EOF) )
  {
    comment = FALSE;
    chptr   = buffer;
    *chptr  = 0;

    commcount = 0;
    comptr    = commenttext;
    *comptr   = 0;

    ch = fgetc(fp);
    while ( (ch != '\n') && (ch != '\r') && (ch != EOF) )
    {

      /* This version of the parser will return an initial character of
         CNF_COMMENT (=';') for FKISS or cel group character,
         followed by the remainder of the line.
      */

      if (ch == CNF_COMMENT)
      { if ( chcount > 0 )
          comment = TRUE;
        else
        {
          ch = fgetc(fp);
          ungetc(ch,fp);
          if ( (ch != CNF_FKISS) && (ch != CNF_CELGROUP) )
            comment = TRUE;
          ch = CNF_COMMENT;
        }
      }  /* if ( ch == CNF_COMMENT ) */

     /* Patch to lose top-bit set chars (Kanji) & controls */

      if ( ch & 0x80 ) ch = ' ';

      if ( ch != '\r' )
      {
        if ( ch == '\t' ) ch = ' ';

        if ( comment )
        {
          if ( commcount < commentlen )
          {
            *comptr = ch;
            comptr++;
            *comptr = 0;
            commcount++;
          }
        }
        else
          if ( chcount < bufflen )
          {
             *chptr = ch;
             chptr++;
             *chptr = 0;
             chcount++;
          }
      } /* if ( ch != 'r' ) */

      ch = fgetc(fp);
    }  /* while ( not end of line or EOF ) */

    if ( ch == '\r' )
    {
      nextch = fgetc(fp);
      if ( nextch != '\n' )
        ungetc(nextch,fp);
      (*lineno)++;
    }
    else
      if ( ch == '\n' )
      {
        nextch = fgetc(fp);
        if ( nextch != '\r' )
          ungetc(nextch,fp);
        (*lineno)++;
      }

    /*
     * Check for total line length
     */

     if ( (chcount + commcount) > MAX_LINE_LENGTH )
       message_error("CNF29");

   /* Strip off trailing spaces */

    while ( (chcount>0) && (chptr[-1] == ' ') )
    { chptr--;
      chcount--;
      *chptr = 0;
    }

   /* Check for FKISS introducer */

    if ( (*buffer == CNF_COMMENT) && (buffer[1] == '@') )
    {
      char *fkissptr = &(buffer[2]);

      skipspace(&fkissptr);

      /* We now need to check to see if this line introduces an FKISS event handler */

      if ( struncmp(fkissptr,FKISS_ENABLE,strlen(FKISS_ENABLE))==0 )
      {
        if ( strcmp(fkissptr,FKISS_ENABLE) == 0 )
        {
          if ( fkiss_flags & FKISS_K_Handler )
            message_error("FKS26");
        }
        else message_error("FKS0");
        fkiss_flags |= FKISS_K_Handler;
        chcount = 0;
      }

     /*
      * Check to see if this /could/ be FKISS, but missing an EventHandler
      */

      if ( chcount > 0 )
      {
        if ( !(fkiss_flags & FKISS_K_Handler) )
        {
          if ( !possible_fkiss_detected && IsFKISSFunction(fkissptr) )
          {
            possible_fkiss_detected = TRUE;
            message_error("FKS27",fkissptr);
          }
          chcount = 0;
        }  /* No EventHandler detected so far */
      }    /* There are non-discarded characters on the line */
    }      /* Line begins ';@', so possible FKISS */

    /*
     * Check for "pragma" i.e. HINT
     */

    if ( comment && (chcount == 0) && (commcount > 0) )
    {
      if ( pragma_setPragma(&(commenttext[1])) )
        commenttext[0] = 0;
    }
  }        /* While we don't have a line to return */

  if ( buffstart ) *buffstart = &(buffer[1]);

  /* Force a space at the end! */

  if ( chcount > 0 )
  {
    *chptr = ' ';
    chptr[1] = 0;
  }

  return chcount;
}

/***************************************************************************/

typedef BOOL (*old_format_fn)(FILE *fp,void *hdr);

BOOL read_header(FILE *fp,void *header,int hdrlen,
                 old_format_fn readold,BOOL *gs2)
{ int newkey;

  if ( gs2 )
    *gs2 = TRUE;
  if ( fread(&newkey,sizeof(int),1,fp) != 1) return FALSE;

#if(BIGENDIAN==1)
  newkey = SWAP_LONG(newkey);
#endif

   if ( newkey == KISS_ID )
    return (BOOL)((fseek(fp,0,SEEK_SET)==0) && (fread(header,hdrlen,1,fp) == 1));

  if ( gs2 )
    *gs2 = FALSE;

  memset(header,0,hdrlen);
  return (BOOL)((fseek(fp,0,SEEK_SET)==0) && readold(fp,header));
}

/***************************************************************************/

KISS_OBJECT_LIST get_object_mark(int mark)
{
  KISS_OBJECT_LIST object = KISS_objects;
  while ( object && (object->mark != mark) ) object = object->next;
  return object;
}

/***************************************************************************/

KISS_CEL_LIST get_cel_descriptor(char *celname)
{
   KISS_CEL_LIST cel = KISS_cels;

  while ( cel )
  {
    if ( strucmp(cel->celname,celname) == 0 )
      return cel;
    cel = cel->next;
  }
  return NULL;
}

