/***************************************************************************/
/*                                                                         */
/*                            KISS archive validator                       */
/*                                                                         */
/***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>

#include "BOOL.h"

#include "KISS_data.h"

#include "KISSTypes.h"
#include "FileHdrTyp.h"

#include "readkiss.h"
#include "message.h"
#include "utilities.h"
#include "osspecific.h"
#include "FKISS_IO.h"
#include "celgroup.h"
#include "pragma.h"

#define CNF_EOF      0x1a

#define ALL_SETS_MASK ((1<<MAX_SETS)-1)

/* Maximum length of a line of KISS configuration */

static int total_colours = 0;
static BOOL non_cherry_cel = FALSE;

int line_count = 0;

/***************************************************************************/
/*
 * Compute either the number of characters yet to read from a file,
 * OR the total number of characters in a file.
 */

static long int fextent(FILE *fp,BOOL total)
{
  long int currpos, endpos;

  currpos = ftell(fp);
  fseek(fp,0,SEEK_END);
  endpos  = ftell(fp);
  fseek(fp,currpos,SEEK_SET);
  return endpos - (total ? 0 : currpos);
}

/***************************************************************************/
/*
 * Set up a header for a KISS/GS1 palette file as it would appear in KISS/GS2
 */

static BOOL make_old_pal_header(FILE *fp,void *h)
{ KISS_PAL_HEADER *pal = h;

  pal->mark        = KISS_PAL_MARK;
  pal->bpcol       = 12;
  pal->colspergrp  = 16;
  pal->numgroup    = 10;
  return TRUE;
}

/***************************************************************************/
/*
 * Validate a KISS palette header.
 */

static BOOL validate_pal_hdr(KISS_PAL_HEADER *pal)
{

#if(BIGENDIAN==1)
  pal->kissdataid = SWAP_LONG(pal->kissdataid);
  pal->colspergrp = SWAP_SHORT(pal->colspergrp);
  pal->numgroup   = SWAP_SHORT(pal->numgroup);
#endif

  if ( pal->mark != KISS_PAL_MARK )
  {
    message_error("CNF3a",pal->mark,KISS_PAL_MARK);
    return FALSE;
  }
  if ( (pal->bpcol != 12) && (pal->bpcol != 24) )
  {
    message_error("CNF3b",pal->bpcol);
    return FALSE;
  }
  if ( !INRANGE(pal->colspergrp,1,PAL_MAX_CPG) )
  {
    message_error("CNF3c",pal->colspergrp);
    return FALSE;
  }
  if ( !INRANGE(pal->numgroup,1,PAL_MAX_GRP) )
  {
    message_error("CNF3d");
    return FALSE;
  }
  return TRUE;
}

/***************************************************************************/

static BOOL validate_palette_body(FILE *fp,KISS_PAL_HEADER *palhdr)
{
  int col, bytestoread;

  /* How many colours will this palette use up? */

  col = palhdr->colspergrp - (PaletteCount>0);

  /* Do we have enough colours left in a standard palette set ? */

  if ( (total_colours + col) > PAL_MAX_CPG )
    KISS_ColourModel = MAX(KISS_ColourModel,KISS_V_EXTENDEDPALETTE);

  /* Update the total colour count */

  total_colours += col;

  /*
   * Compare the actual length of the file with the number of bytes
   * left to read (i.e. is the file readable) ?
   */

  bytestoread = (palhdr->bpcol == 12) ? 2 : 3;
  bytestoread *= palhdr->colspergrp * palhdr->numgroup;

  PaletteCount++;

  return (BOOL)(fextent(fp,FALSE) >= bytestoread);
}

/***************************************************************************/

static void addPaletteFile(char *filename,int colspergrp)
{
  KISS_PALETTE_LIST pal = safe_alloc(sizeof(KISS_PALETTE_TYPE));

  pal->name         = strdup(filename);
  pal->colspergroup = colspergrp;
  pal->next         = NULL;

  if ( KISS_palettes == NULL )
    KISS_palettes = pal;
  else
  {
    KISS_PALETTE_LIST lastpal = KISS_palettes;
    while ( lastpal && lastpal->next )
      lastpal = lastpal->next;
    lastpal->next = pal;
  }
}

/***************************************************************************/

static KISS_PALETTE_LIST lookupPaletteNumber(int palno)
{
  KISS_PALETTE_LIST pal = KISS_palettes;

  while ( pal && (palno > 0) )
  {
    palno--;
    pal = pal->next;
  }
  return pal;
}

/***************************************************************************/
/*
 * Validate a complete palette file
 */

static BOOL validatePaletteFile(char *filename)
{
  FILE *pf;
  BOOL success,gs2;
  KISS_PAL_HEADER palhdr;

  pf = fopen_path(filename,"r");

  success = pf != NULL;
  if ( pf )
  {
    success = read_header(pf,&palhdr,sizeof(KISS_PAL_HEADER),make_old_pal_header,&gs2)
                 && validate_pal_hdr(&palhdr);
    if ( success )
    {
      /* Read the palette file */

      success = validate_palette_body(pf,&palhdr);
      if ( !success )
        message_error("CNF3e");
      if ( gs2 )
        KISS_ColourModel = MAX(KISS_ColourModel,KISS_V_GS2PALETTE);
      else KISS_ColourModel = MAX(KISS_ColourModel,KISS_V_GS1PALETTE);
      addPaletteFile(filename,palhdr.colspergrp);
    }

    fclose(pf);
  }
  else message_error("CNF2",filename);
  return success;
}

/***************************************************************************/

static BOOL make_old_cel_header(FILE *fp,void *h)
{
  KISS_CEL_HEADER *cel = h;

  cel->mark = KISS_CEL_MARK;
  cel->bpp  = 4;
  return (BOOL)( (fread(&(cel->width),sizeof(short int),1,fp) == 1) &&
                 (fread(&(cel->height),sizeof(short int),1,fp) == 1) );
}

/***************************************************************************/

static BOOL validate_cel_header(KISS_CEL_HEADER *cel)
{
  if ( cel->mark == KISS_CEL_MARK )
  {
    if ( (cel->bpp !=4) && (cel->bpp != 8) )
    {
      message_error("CNF8a");
      return FALSE;
    }
    non_cherry_cel = TRUE;
  }
  else
    if ( cel->mark == KISS_C32_MARK )
    {
      if ( cel->bpp != 32 )
      {
        message_error("CNF8b");
        return FALSE;
      }
      CherryKISS = TRUE;
      KISS_ColourModel = MAX(KISS_ColourModel,KISS_V_CHERRY);
    }
    else
    {
      message_error("CNF8c",cel->mark,KISS_CEL_MARK,KISS_C32_MARK);
      return FALSE;
    }
  return TRUE;
}

/***************************************************************************/

static BOOL validate_cel_body(FILE *fp,KISS_CEL_HEADER *hdr)
{
  int bytesperrow, totalbytes;

  switch ( hdr->bpp )
  {
    case 4 : bytesperrow = (hdr->width +1)>>1;
             break;
    case 8 : bytesperrow = hdr->width;
             break;
    case 32: bytesperrow = hdr->width * 4;
  }

  totalbytes = bytesperrow * hdr->height;
  return (BOOL)(fextent(fp,FALSE) >= totalbytes);
}

/***************************************************************************/

static BOOL findObject(int               mark,
                       KISS_OBJECT_LIST *object,
                       KISS_OBJECT_LIST *prev)
{
  KISS_OBJECT_LIST obj=KISS_objects,p=NULL;

  while ( obj && (obj->mark < mark) )
  {
    p   = obj;
    obj = obj->next;
  }

  if ( prev )
    *prev = p;
  if ( object )
    *object = obj;

  return (BOOL)(obj && (obj->mark == mark));
}

/***************************************************************************/

static KISS_OBJECT_LIST findOrCreateObject(int mark)
{
  KISS_OBJECT_LIST object, prev, newobj;

  if ( findObject(mark,&object,&prev) )
    return object;

  newobj = safe_alloc(sizeof(KISS_OBJECT_TYPE));
  memset(newobj,0,sizeof(KISS_OBJECT_TYPE));

  newobj->next = object;
  if ( prev )
       prev->next = newobj;
  else KISS_objects = newobj;

  newobj->mark = mark;

  ObjectCount++;

  return newobj;
}

/***************************************************************************/

static KISS_OBJECT_LIST add_cel_to_object(KISS_CEL_LIST cel,int mark)
{
  KISS_OBJECT_LIST object;
  KISS_CEL_LIST objcel, dupcel;

  object  = findOrCreateObject(mark);

  /*
   * Check to see if this object already uses this cel
   */

  dupcel = cel->duplicate;
  objcel = object->first_cel;

  object->fix = MAX(object->fix,cel->fix);

  cel->objectid = object;

  cel->next_cel     = object->first_cel;
  object->first_cel = cel;

  object->size.x    = MAX(object->size.x,(cel->offset.x+cel->size.x));
  object->size.y    = MAX(object->size.y,(cel->offset.y+cel->size.y));

  return object;
}

/***************************************************************************/

static BOOL validateCelFile(char          *srcfile,
                            KISS_CEL_TYPE *cel,
                            KISS_CEL_LIST *existingcel)
{
  FILE *pf;
  KISS_CEL_HEADER hdr;
  BOOL gs2, valid;

  *existingcel = NULL;

  pf = fopen_path(srcfile,"rb");

  if ( !pf )
  {
    message_error("CNF7",srcfile);
    return FALSE;
  }

  valid = read_header(pf,&hdr,sizeof(KISS_CEL_HEADER),make_old_cel_header,&gs2);
  if (  valid && validate_cel_header(&hdr) )
  {
    /* Compute the required object size for this cel */

#if(BIGENDIAN==1)
    hdr.width    = SWAP_SHORT(hdr.width);
    hdr.height   = SWAP_SHORT(hdr.height);
    hdr.x_offset = SWAP_SHORT(hdr.x_offset);
    hdr.y_offset = SWAP_SHORT(hdr.y_offset);
#endif

    cel->size.x   = hdr.width;
    cel->size.y   = hdr.height;
    cel->offset.x = hdr.x_offset;
    cel->offset.y = hdr.y_offset;
    cel->bpp      = hdr.bpp;
    if ( cel->bpp > 8 )
      cel->flags |= CEL_K_Cherry;

    *existingcel   = get_cel_descriptor(srcfile);

    if ( *existingcel == NULL )
    {
      if ( !validate_cel_body(pf,&hdr) )
      {
        fclose(pf);
        message_error("CNF9",srcfile);
        return FALSE;
      }
    }

    if ( gs2 )
      KISS_Version = MAX(KISS_Version,KISS_V_GS2);
  }
  else
  {
    fclose(pf);
    message_error("CNF8",srcfile);
    return FALSE;
  }

  fclose(pf);
  return TRUE;
}

/***************************************************************************/
/*
 * Add the cel to the list of cels, if necessary
 */

static KISS_CEL_LIST addCelToList(KISS_CEL_TYPE *cel,
                                 KISS_CEL_LIST existingcel)
{
  KISS_CEL_LIST newcel;

  /* Now create the cel */

  newcel = safe_alloc(sizeof(KISS_CEL_TYPE));
  memcpy(newcel,cel,sizeof(KISS_CEL_TYPE));

  newcel->next = NULL;
  newcel->next_cel     = NULL;

  if ( KISS_cels )
  {
    newcel->prev = KISS_cels->prev;
    KISS_cels->prev->next = newcel;
    KISS_cels->prev = newcel;
  }
  else
  {
    newcel->prev = newcel;
    KISS_cels = newcel;
  }

  /* And set up the duplicate cel link */

  while ( existingcel && (existingcel->duplicate) )
    existingcel = existingcel->duplicate;

  if ( existingcel )
    existingcel->duplicate = newcel;

  return newcel;
}

/***************************************************************************/

static BOOL attrFlagSet(KISS_CEL_LIST cel,unsigned int flag)
{
  BOOL alreadySet=FALSE;
  if ( cel->flags & flag )
  {
    message_error("CNF30");
    alreadySet = TRUE;
  }
  else cel->flags |= flag;

  return !alreadySet;
}

/***************************************************************************/

static void processAttributes(KISS_CEL_LIST cel,
                              char          **commentstring,
                              BOOL          basicAllowed)
{
  char *comment = *commentstring;
  char grpname[MaxIdentifierLen], term[10], *p;
  int valread;

  if ( basicAllowed )
  {
    sprintf(term," %c%c",CNF_CELATTR,CNF_CELGROUP);

    if ( *comment != CNF_CELATTR )
    {
      skipspace(&comment);
      if ( *comment == CNF_CELATTR )
        message_error("CNF36");
    }

    while ( *comment == CNF_CELATTR )
    {
      comment++;
      switch ( *comment )
      {
        case 't' :
          comment++;

          if ( cel->starttransp > 0 )
            message_error("CNF30");
          else
            if ( !isdigit(*comment) || !
                 read_integer(&comment,term,&valread) )
              message_error("CNF20");
            else
              cel->starttransp = valread;
          break;

        case 'g' :
          comment++;
          attrFlagSet(cel,CEL_K_StartGhosted);
          KISS_Version = MAX(KISS_Version,KISS_V_FKISS4);
          break;

        case 'u' :
          comment++;
          attrFlagSet(cel,CEL_K_StartUnmapped);
          KISS_Version = MAX(KISS_Version,KISS_V_FKISS4);
          break;

        case 'x' :
        case 'y' :
          {
            unsigned int flag = tolower(*comment) == 'x' ? CEL_K_OverrideXOff
                                                         : CEL_K_OverrideYOff;
            comment++;
            if ( !isdigit(*comment) || !read_integer(&comment,term,&valread) )
            {
              message_error("CNF36");
            }
            else
              if ( attrFlagSet(cel,flag) )
              {
                if ( flag == CEL_K_OverrideXOff )
                   cel->offset.x = valread;
                else cel->offset.y = valread;
              }
          }
          KISS_Version = MAX(KISS_Version,KISS_V_FKISS4);
          break;

        default :
          message_error("CNF31",*comment);
          comment++;
          break;

      } /* switch */
    skipspace(&comment);
    }   /* while */
  }     /* if */

  /*
   * Read any cel groups/frames for this cel
   */

  sprintf(term," ;%c",CNF_CELGROUP);

  while ( *comment == CNF_CELGROUP )
  {
    comment++;
    skipspace(&comment);
    if ( isalpha(*comment) )
    {
      p       = grpname;
      valread = 0;
      while ( isalnum(*comment) && (valread < MaxIdentifierLen) )
      {
        *p = *comment;
        p++;
        comment++;
        valread++;
      }

      if ( isalnum(*comment) )
      {
        message_error("CNF31");
        while ( isalnum(*comment) ) comment++;
      }
      else *p = 0;

      skipspace(&comment);
      if ( *comment != CNF_FRAMEBEGIN )
        celgroup_addCelToGroup(grpname,NULL_FRAME,cel);
      else
      {
        comment++;
        skipspace(&comment);
        if ( !isdigit(*comment) )
          message_error("CNF32");
        while ( isdigit(*comment) )
        {
          if ( !read_integer(&comment,term,&valread) )
          {
            message_error("CNF32");
            break;
          }
          celgroup_addCelToGroup(grpname,valread,cel);
          skipspace(&comment);
        }
#ifdef CNF_FRAMEEND
        if ( *comment != CNF_FRAMEEND )
          message_error("CNF35");
#endif
      }
    }
    else
    {
      message_error("CNF33");
      break;
    }
  }
  *commentstring = comment;
}

/***************************************************************************/

static KISS_CEL_LIST read_cel(char **cnf_ptr, int celno,int palcnt,char *comment)
{
  char filename[MAXLEAFNAMELEN];
  int filelen,mark = 0;
  KISS_CEL_LIST newcel, existingcel;
  KISS_CEL_TYPE cel;
  KISS_OBJECT_LIST mark_object;
  char *celterm;

  memset(&cel,0,sizeof(KISS_CEL_TYPE));

  /* Read Mark and Fix */

  if ( !read_integer(cnf_ptr,". ",&mark) )
  {
    message_error("CNF12");
    return NULL;
  }

  skipspace(cnf_ptr);

  if ( **cnf_ptr == '.' )
  { (*cnf_ptr)++;
    if ( !isspace(**cnf_ptr) )
    {
      if ( !read_integer(cnf_ptr," ",&cel.fix) )
      {
        message_error("CNF18");
        cel.fix = 0;
      }
      else
        if ( cel.fix > MAXFIXVAL )
        {
          message_error("CNF18");
          cel.fix = MAXFIXVAL;
        }
    }
  }

  memset(filename,0,MAXLEAFNAMELEN);
  skipspace(cnf_ptr);

  if ( (**cnf_ptr == '\'') || (**cnf_ptr == '\"') )
  {
    message_error("CNF24");
    celterm = " *:\'\"";
    (*cnf_ptr)++;

  }
  else celterm = " *:";

  if ( !read_next_field(cnf_ptr,celterm,filename,MAXLEAFNAMELEN,&filelen) )
  {
    message_error("CNF21");
    return NULL;
  }

  if ( (**cnf_ptr == '\'') || (**cnf_ptr == '\"') )
    (*cnf_ptr)++;

  if ( strlen(filename) > MAXDOSNAMELEN )
    message_error("CNF22",filename);

  skipspace(cnf_ptr);

  if ( **cnf_ptr == '*' )
  {
    (*cnf_ptr)++;
    if ( !read_integer(cnf_ptr," :",&cel.palette) )
    {
      message_error("CNF17");
      cel.palette = 0;
    }
    skipspace(cnf_ptr);
  }

  /* Set up the set mask */

  if ( **cnf_ptr == ':' )
  {
    int setno=0;
    BOOL errored = FALSE;

    (*cnf_ptr)++;
    skipspace(cnf_ptr);
    while (**cnf_ptr)
    {
      if ( !read_integer(cnf_ptr," ",&setno) || (setno>MAX_SETS) )
      {
        message_error("CNF19");
        errored = TRUE;
        break;   /* A non-integer just makes us ignore the rest of the line */
      }

      setno = 1 << setno;
      cel.setmask |= setno;
      skipspace(cnf_ptr);
    }
    if ( (cel.setmask == 0) && !errored )
    {
      message_error("CNF23");
      cel.setmask = ALL_SETS_MASK;
    }
  }
  else cel.setmask = ALL_SETS_MASK;

  ActiveSetsMask |= cel.setmask;

  /* Now create the cel & object entries in the data table */

  if ( syntax_only )
  {
    if ( lzhArchiveContents && !lzh_findFile(lzhArchiveContents,filename) )
      message_error("CNF7",filename);

    existingcel = get_cel_descriptor(filename);
  }
  else
  {
    if ( !validateCelFile(filename,&cel,&existingcel) )
      return NULL;

    /*
     * Check for valid palette
     */

    if ( cel.bpp < 16 )
    {
      if (cel.palette >= palcnt)
        message_error((palcnt==0)?"CNF17a":"CNF17b",filename);
      else
      {
        KISS_PALETTE_LIST pal = lookupPaletteNumber(cel.palette);
        if ( pal->colspergroup != (1u<<cel.bpp) )
          message_error("CNF17c",filename,pal->name);
      }
    }
  }

  /* Cel loaded OK. Now turn it into an entry on the list */

  cel.celname = strdup(filename);

  if ( newcel = addCelToList(&cel,existingcel), newcel == NULL )
    return NULL;

  if ( *comment )
  {
    comment++;
    processAttributes(newcel,&comment,TRUE);
  }

  mark_object = add_cel_to_object(newcel,mark);

  MaxFixValue = MAX(cel.fix,MaxFixValue);

  return newcel;
}

/***************************************************************************/

static BOOL object_in_set(KISS_OBJECT_LIST obj,int setno)
{
  KISS_CEL_LIST cel = obj->first_cel;

  setno = (1u<<setno);

  while ( cel )
  {
    if ( cel->setmask & setno )
      return TRUE;

    cel = cel->next_cel;
  }

  return FALSE;
}

/***************************************************************************/

static BOOL process_set(int setno,char **cnfptr,int *objid)
{ int x,y;
  KISS_OBJECT_LIST object;

  skipspace(cnfptr);
  while ( **cnfptr )
  {
    if ( **cnfptr == '*' )
    {
      (*cnfptr)++;

     /* Check to see if a null coordinate is sensible */

      object = get_object_mark(*objid);
      if ( object )
      {
        if ( object_in_set(object,setno) )
          message_error("CNF25",*objid,setno);
      }
    }
    else
    {
      if ( !read_integer(cnfptr,",",&x) )
        return FALSE;

      (*cnfptr)++;
      if ( !read_integer(cnfptr," ,",&y) )
        return FALSE;

      if ( **cnfptr != ' ' )
        return FALSE;

      object = get_object_mark(*objid);
      if ( object )
      {
        object->startpos[setno].x = x;
        object->startpos[setno].y = y;
        object->setmask |= 1<<setno;

        if ( !object_in_set(object,setno) )
          message_error("CNF26",*objid,setno);
      }
      else message_error("CNF27",*objid,setno);
    }
    (*objid)++;
    skipspace(cnfptr);
  }

  return TRUE;
}

/***************************************************************************/

static BOOL checkGSRank(int rank)
{
  return ( (imagesize.x <= rankLimits[rank].maxSize.x) &&
           (imagesize.y <= rankLimits[rank].maxSize.y) &&
           (total_colours <= rankLimits[rank].maxColours) &&
           (CelCount <= rankLimits[rank].maxCels) &&
           (ObjectCount <= rankLimits[rank].maxObjects) );
}

/***************************************************************************/

static BOOL validate_doll(FILE *fp, char *cnf)
{ char cnf_buffer[MAX_INPUT_BUFFER_LEN+4],
       comment[MAX_INPUT_BUFFER_LEN+4], *cnf_ptr;
  int line_len;
  int setcnt = 0;
  volatile int status;
  BOOL bad_text_error = FALSE;
  KISS_CEL_LIST last_cel = NULL, current_cel = NULL;

  line_count = 0;

  line_len =  read_line(fp,cnf_buffer,MAX_INPUT_BUFFER_LEN,
                           comment,MAX_INPUT_BUFFER_LEN,&line_count,&cnf_ptr);

  status = setjmp(read_doll_env);
  if ( status == 0 )
  {
    while ( line_len > 0 )
    { BOOL line_pending = FALSE;

      current_cel = last_cel;
      last_cel    = NULL;

      switch ( *cnf_buffer )
      {
         case CNF_MEMORY  : break;

         case CNF_SCRNSZ  : /* Note we allow several options for the separator */
          { COORD_PAIR imgsz;
            if ( !read_integer(&cnf_ptr,",.*xX",&(imgsz.x)) )
            {
              message_error("CNF10");
              break;
            }
            if ( *cnf_ptr != ',' )
            {
              message_error("CNF10");
              break;
            }
            cnf_ptr++;
            if ( !read_integer(&cnf_ptr,") ",&(imgsz.y)) )
            {
              message_error("CNF10");
              break;
            }
            if ( *cnf_ptr != ')' )
              message_error("CNF10b");
            requestedSize = imgsz;
            imagesize     = imgsz;
          }
          break;


         case CNF_PALETTE :
            { char palfilename[MAXLEAFNAMELEN];
              int filelen;
              if ( !read_next_field(&cnf_ptr," ",palfilename,MAXLEAFNAMELEN,&filelen) )
                message_error("CNF4");
              else
              {
                if ( strlen(palfilename) > MAXDOSNAMELEN )
                  message_error("CNF22",palfilename);
                if ( !syntax_only )
                {
                  if ( !validatePaletteFile(palfilename) )
                    message_error("CNF3",palfilename);
                }
                else
                {
                  if ( lzhArchiveContents && !lzh_findFile(lzhArchiveContents,palfilename) )
                    message_error("CNF3f",palfilename);
                  else
                  {
                    PaletteCount++;
                    addPaletteFile(palfilename,0);
                  }
                }
              }
            }
            break;

         case CNF_BORDER  :
           /* Some KISS files include the border code, but do not give a digit.
              We assume this forces them to default to zero. */

            if ( !read_integer(&cnf_ptr," ",&border_colour) )   /* ????? Should be an error */
              border_colour = 0;
            break;

         case CNF_CEL     :
            last_cel = read_cel(&cnf_ptr,CelCount,PaletteCount,comment);
            if ( last_cel )
              CelCount++;
            break;

         case CNF_SET     :
            if ( setcnt >= MAX_SETS )
            { message_error("CNF6");
              longjmp(read_doll_env,1);
            }
            else
            { int objid = 0;

              if ( (!read_integer(&cnf_ptr," ",&(KISS_palette_group[setcnt]))) ||
                   (KISS_palette_group[setcnt] < 0) ||
                   (KISS_palette_group[setcnt] >= PAL_MAX_GRP) )
              {
                KISS_palette_group[setcnt] = 0;
                message_error("CNF13",setcnt);
              }

              if ( !process_set(setcnt,&cnf_ptr,&objid) )
                message_error("CNF11",setcnt);

              /* Read a possible continuation line */
              line_len = read_line(fp,cnf_buffer,MAX_INPUT_BUFFER_LEN,
                                      comment,MAX_INPUT_BUFFER_LEN,
                                      &line_count,&cnf_ptr);

              while ( (*cnf_buffer == CNF_SET_CONT) && (line_len > 0) )
              {
                if ( !process_set(setcnt,&cnf_ptr,&objid) )
                  message_error("CNF11",setcnt);

                /* Read a set continuation */

                line_len = read_line(fp,cnf_buffer,MAX_INPUT_BUFFER_LEN,
                                        comment,MAX_INPUT_BUFFER_LEN,
                                        &line_count,&cnf_ptr);
              }
              line_pending = TRUE;
              setcnt++;

            }
            break;

         case CNF_EOF : break;

         case CNF_COMMENT :
             /* This is now legal, iff the next character is CNF_FKISS */
             switch (*cnf_ptr)
             {
               case CNF_FKISS :
                     process_fkiss_command_line(&(cnf_ptr[1]));
                     break;

               case CNF_CELGROUP :
                     if ( current_cel )
                     {
                       processAttributes(current_cel,&cnf_ptr,FALSE);
                       last_cel = current_cel;
                     }
                     else
                       message_error("CNF34");
                     break;

               default : break;
             }
             break;

         default     :
             if ( !bad_text_error )
             { message_error("CNF0");
               bad_text_error = TRUE;
             }
             break;
      }
      if ( !line_pending )
        line_len = read_line(fp,cnf_buffer,MAX_INPUT_BUFFER_LEN,
                                comment,MAX_INPUT_BUFFER_LEN,
                                &line_count,&cnf_ptr);
    }

    longjmp(read_doll_env,1);
  }

  /*
   * If there are any non-Cherry cels in a cherry set,
   * treat it as extended palette
   */

  if ( CherryKISS && non_cherry_cel )
    KISS_ColourModel = KISS_V_HYBRID;

  if ( KISS_ColourModel > KISS_V_GS1PALETTE )
    KISS_Version = MAX(KISS_Version,KISS_V_GS2);

  if ( (KISS_Version < KISS_V_GS2PLUS) &&
       !checkGSRank(KISS_Version) )
    KISS_Version = (KISS_Version == KISS_V_GS1) ? KISS_V_GS1PLUS : KISS_V_GS2PLUS;

  return (status >= 0);
}

/***************************************************************************/

static void deleteDoll(void)
{
  KISS_PALETTE_LIST palette;
  KISS_OBJECT_LIST  obj;
  KISS_CEL_LIST     cel;

  /*
   * Delete the palette, object and cel lists
   */

  while ( KISS_palettes )
  {
    palette       = KISS_palettes;
    KISS_palettes = KISS_palettes->next;
    free(palette->name);
    free(palette);
  }

  while ( KISS_objects )
  {
    obj          = KISS_objects;
    KISS_objects = KISS_objects->next;
    free(obj);
  }

  while ( KISS_cels )
  {
    if ( KISS_cels->celname )
      free(KISS_cels->celname);
    cel       = KISS_cels;
    KISS_cels = KISS_cels->next;
    free(cel);
  }

  /*
   * Junk any cel groups and hints
   */

  celgroup_dispose();
  pragma_reset();

  /*
   * Junk all FKISS scripts
   */

  fkiss_clear_doll();

  /*
   * Re-initialize all the state variables
   */

  non_cherry_cel     = FALSE;
  total_colours      = 0;
  MaxFixValue        = 0;
  CherryKISS         = FALSE;
  ExtendedPalette    = FALSE;
  ActiveSetsMask     = 0;
  ObjectCount        = 0;
  CelCount           = 0;
  PaletteCount       = 0;
  line_count         = 0;
  imagesize.x        = 0;
  imagesize.y        = 0;
  requestedSize.x    = 0;
  requestedSize.y    = 0;

  fkiss_flags        = FKISS_K_Initial;
  KISS_Version       = KISS_V_GS1;
  KISS_ColourModel   = KISS_V_NOMODEL;
}

/***************************************************************************/

BOOL KISS_validateDoll(char *cnf)
{
  FILE *fp;
  BOOL validated_ok;

  deleteDoll();

  if ( fp = fopen_cnf(cnf,"r"), fp == NULL )
  {
    message_error("CNF2",cnf);
    return FALSE;
  }

  message_begin(cnf);

  /* Now load the doll */

  validated_ok = validate_doll(fp,cnf);

  fclose(fp);

  fkiss_load_completed();

  return validated_ok;
}

